import os
import time
import globus_sdk
import urllib
import logging
from fair_research_login import NativeClient, LoadError, ScopesMismatch
from pilot import (
    profile, config, globus_clients, exc, logging_cfg, context, search,
    transfer_log, search_discovery, project as project_module,
)

logging_cfg.setup_logging()
log = logging.getLogger(__name__)


class PilotClient(NativeClient):
    r"""
    The Pilot Client is the core class of this library, providing a set of
    operations to authenticate users, track which project they're using, and
    handle core operations such as login, logout, upload, download,
    building URLs, and getting authenticated Globus Service clients.

    ** Methods **

    *  :py:meth:`.login`
    *  :py:meth:`.logout`
    *  :py:meth:`.is_logged_in`
    *  :py:meth:`.get_auth_client`
    *  :py:meth:`.get_search_client`
    *  :py:meth:`.get_transfer_client`
    *  :py:meth:`.get_http_client`
    *  :py:meth:`.get_group`
    *  :py:meth:`.get_endpoint`
    *  :py:meth:`.get_index`
    *  :py:meth:`.get_path`
    *  :py:meth:`.get_globus_http_url`
    *  :py:meth:`.get_globus_url`
    *  :py:meth:`.get_globus_app_url`
    *  :py:meth:`.get_portal_url`
    *  :py:meth:`.get_subject_url`
    *  :py:meth:`.ls`
    *  :py:meth:`.mkdir`
    *  :py:meth:`.get_search_entry`
    *  :py:meth:`.ingest`
    *  :py:meth:`.ingest_many`
    *  :py:meth:`.ingest_gmeta`
    *  :py:meth:`.delete_entry`
    *  :py:meth:`.upload`
    *  :py:meth:`.download`
    *  :py:meth:`.delete`
    ** Example **
    Login, upload, show various access methods, then delete.

    pc = PilotClient()

    # Setup the client (You can also do this with the CLI, both use the same
    #                   credentials)
    pc.login()
    pc.project.current = 'foo'

    # Create a directory, Upload a file to it
    pc.mkdir('bar')
    pc.upload('moo.txt', 'bar')

    # Get resources associated with the file
    pc.ls('bar')
    pc.get_path('bar/moo.txt')
    pc.get_portal_url('bar/moo.txt')
    pc.get_globus_app_url('bar/moo.txt')
    pc.get_globus_http_url('bar/moo.txt')
    pc.download('bar/moo.txt')

    # Delete the file
    pc.delete_entry('bar/moo.txt')
    pc.delete('bar/moo.txt')
    """
    DEFAULT_SCOPES = [
        'profile',
        'openid',
        'urn:globus:auth:scope:search.api.globus.org:all',
        'urn:globus:auth:scope:transfer.api.globus.org:all',
    ]
    GROUPS_SCOPE = 'urn:globus:auth:scope:nexus.api.globus.org:groups'
    DISALLOWED_FILENAME_SYMBOLS = '.*~$%'
    DEFAULT_CONFIG = '~/.pilot1.cfg'

    def __init__(self, config_file=DEFAULT_CONFIG, index_uuid=None):
        # Supplying config by Env is strongest and overrides all others
        # Precedence: ENV --> __init__ Argument --> DEFAULT_CONFIG
        self.config_file = config_file
        env_conf = os.getenv('PILOT_CONFIG')
        if env_conf:
            self.config_file = env_conf
            log.info(f'Pilot Config ENV overrided by {env_conf}')
        if config_file is not None:
            self.config_file = os.path.expanduser(self.config_file)
        log.debug(f'Set config to {self.config_file}')
        self.config = config.Config(self.config_file)
        self.context = context.Context(self, config=self.config,
                                       index_uuid=index_uuid)
        default_scopes = self.context.get_value('scopes')
        default_scopes = default_scopes or self.DEFAULT_SCOPES

        super().__init__(client_id=self.context.get_value('client_id'),
                         token_storage=self.config,
                         default_scopes=default_scopes,
                         app_name=self.context.get_value('app_name'))
        self.project = project_module.Project(config=self.config)
        self.profile = profile.Profile(config=self.config)
        self.transfer_log = transfer_log.TransferLog(config=self.config)

    def login(self, *args, **kwargs):
        r"""
        Do a Native App Auth Flow to get tokens for requested scopes, for
        general project usage
        **Parameters**
        ``no_local_server`` (*bool*)
          Disable spinning up a local server to automatically copy-paste the
          auth code. THIS IS REQUIRED if you are on a remote server, as this
          package isn't able to determine the domain of a remote service. When
          used locally with no_local_server=False, the domain is localhost with
          a randomly chosen open port number.
        ``no_browser`` (*string*)
          Do not automatically open the browser for the Globus Auth URL.
          Display the URL instead and let the user navigate to that location.
        ``requested_scopes`` (*list*)
          A list of scopes to request of Globus Auth during login.
          Example:
          ['openid', 'profile', 'email']
        ``refresh_tokens`` (*bool*)
          Ask for Globus Refresh Tokens to extend login time.
        ``force`` (*bool*)
          Force a login flow, even if loaded tokens are valid.
        """
        super().login(*args, **kwargs)
        auth_cli = self.get_auth_client()
        user_info = auth_cli.oauth2_userinfo()
        self.profile.save_user_info(user_info.data)

    def logout(self):
        """Clear tokens from configfile"""
        super().logout()
        self.config.clear()

    def is_logged_in(self):
        """Check if the user is logged in and tokens are active."""
        try:
            self.load_tokens(requested_scopes=self.context.get_value('scopes'))
            return True
        except LoadError:
            return False

    def get_auth_client(self):
        """
        Returns a live Globus Auth Client based on user login info.
        https://globus-sdk-python.readthedocs.io/en/stable/clients/auth/
        :return:
        """
        authorizer = self.get_authorizers()['auth.globus.org']
        return globus_sdk.AuthClient(authorizer=authorizer)

    def get_search_client(self):
        """Returns a live Search Client based on user login info
        https://globus-sdk-python.readthedocs.io/en/stable/clients/search/
        """
        authorizer = self.get_authorizers()['search.api.globus.org']
        return globus_sdk.SearchClient(authorizer=authorizer)

    def get_transfer_client(self):
        """
        Returns a live transfer client based on user info
        https://globus-sdk-python.readthedocs.io/en/stable/clients/transfer/
        """
        authorizer = self.get_authorizers()['transfer.api.globus.org']
        return globus_sdk.TransferClient(authorizer=authorizer)

    def get_nexus_client(self):
        """
        Returns a nexus client, used for groups. Please don't use this directly
        """
        rs = {'requested_scopes': [self.GROUPS_SCOPE]}
        try:
            authorizer = self.get_authorizers_by_scope(**rs)[self.GROUPS_SCOPE]
            return globus_clients.NexusClient(authorizer=authorizer)
        except ScopesMismatch:
            return None

    def get_http_client(self, project):
        r"""
        Returns a general Globus HTTP Client for http uploads/downloads
        **Parameters**
        ``project`` (*bool*)
          The project to upload/download to/from
        """
        # Fetch the base url for the globus http endpoint
        url = urllib.parse.urlparse(self.get_globus_http_url(''))
        base_url = urllib.parse.urlunparse((url.scheme, url.netloc, '', '',
                                            '', ''))
        rs = self.project.get_info(project)['resource_server']
        auth = self.get_authorizers()[rs]
        return globus_clients.HTTPFileClient(authorizer=auth,
                                             base_url=base_url)

    def get_group(self, project=None):
        """
        Get the group for a given project.
        **Parameters**
        ``project`` (*string*)
          The project to fetch info for. Defaults to current project
        """
        return self.project.get_info(project)['group'] or 'public'

    def get_endpoint(self, project=None):
        """
        Get the configured Globus Endpoint for the given project. Project
        defaults to current project if None.
        **Parameters**
        ``project`` (*string*)
          The project to fetch info for. Defaults to current project
        """
        project = project or self.project.current
        return self.project.get_info(project)['endpoint']

    def get_project(self, project=None):
        return self.project.get_info(project or self.project.current)

    def get_context(self, context=None):
        return self.context.get_context(context or self.context.current)

    def get_index(self, project=None):
        """
        Get the configured search index for the given project. Project defaults
        to current project if None.
        **Parameters**
        ``project`` (*string*)
          The project to fetch info for. Defaults to current project
        """
        return self.project.get_info(project)['search_index']

    def get_valid_dataframe(self, dataframe):
        if not dataframe:
            raise exc.InvalidDataframeName(fmt=[dataframe])
        if not isinstance(dataframe, str):
            raise exc.InvalidDataframeName()
        if any([dataframe.startswith(s)
               for s in list(self.DISALLOWED_FILENAME_SYMBOLS)]):
            raise exc.InvalidDataframeName(fmt=[dataframe])
        return os.path.abspath(dataframe)

    def resolve_endpoint(self, url):
        """
        Given an http url or subject, resolve the endpoint within the URL.
        Supports pertel style url suffix with '.e.globus.org'.
        Raises PilotInvalidProject if protocol is not globus, http, or https
        Example URLS:
            globus://foo-endpoint/foo_folder/test_path
            https://foo-endpoint/foo_folder/test_path
            https://foo-endpoint.e.globus.org/foo_folder/test_path
        **Parameters**
        ``url`` (*string*)
          The URL To resolve. Cannot be a short path or fullpath, or None will
          be returned.
        """
        ep = None
        purl = urllib.parse.urlparse(url)
        if purl.scheme not in ['globus', 'http', 'https', '']:
            raise exc.PilotInvalidProject('Invalid protocol '
                                          '{}'.format(purl.scheme))
        if purl.scheme == 'globus':
            ep = purl.netloc
        elif purl.scheme in ['http', 'https']:
            if purl.netloc.endswith('.e.globus.org'):
                ep = purl.netloc.replace('.e.globus.org', '')
        return ep

    def resolve_context(self, url):
        """
        Given a URL, resolve the context to which it belongs. This only works
        for a given context that has a unique endpoint associated with the URL.
        If multiple contexts have the same url, the first context that matches
        will be returned.
        **Parameters**
        ``url`` (*string*)
          The URL To resolve. Cannot be a short path or fullpath, or None will
          be returned.
        """
        ep = self.resolve_endpoint(url)
        if ep:
            for name, cdata in self.context.load_all().items():
                if cdata['projects_endpoint'] == ep:
                    cdata['name'] = name
                    log.debug('Resolved {} to context {}'.format(url, name))
                    return cdata
        log.debug('Failed to resolve context {}'.format(url))
        return None

    def resolve_project(self, url):
        """
        Given a URL, resolve the project to which it belongs. Returns a dict
        containing info about the project.
        **Parameters**
        ``url`` (*string*)
          The URL To resolve. Cannot be a short path or fullpath, or None will
          be returned.
        """
        ep, path = self.resolve_endpoint(url), urllib.parse.urlparse(url).path
        if ep and path:
            for name, pdata in self.project.load_all().items():
                if pdata['endpoint'] == ep and pdata['base_path'] in path:
                    pdata['name'] = name
                    log.debug('Resolved {} to project {}'.format(
                        url, pdata['title']))
                    return pdata
        log.debug('Failed to resolve project {}'.format(url))
        return None

    def build_short_path(self, dataset, destination, project=None):
        """
        Given a source and destination, build the resulting short_path which
        pilot will henceforth use to refer to the remote dataset.

        NOTE: The destination may not exist on the remote endpoint. This only
        give you a preview of what a short_path will look like after an
        operation like ``.upload(dataset, destination)``
        **Parameters**
        ``dataset`` (*path string*)
          A local path to a file or directory
        ``destination`` (*path string*)
          The desired remote location of the dataset.
        """
        sp = os.path.join(destination, os.path.basename(dataset.rstrip('/')))
        return self.get_short_path(sp, project)

    def get_short_path(self, url, project=None):
        """Given a globus HTTP URL, Globus Search subject URL, Globus URL,
        full path or short_path, resolve the short_path for a given project.
        Will raise PilotInvalidProject exception if the url resolved does not
        match the project provided or the current project (if None is given).
        If the base path cannot be resolved, it's assumed the given path was
        a short path and is returned without exception.
        **Parameters**
        ``url`` (*url string*)
          A url to parse into a short path. Pilot mostly uses shortpaths to
          do various operations, relying on the context of the project to save
          the basepath, endpoint, and search index.
        ``project`` (*string* or None)
          Project to use instead of the current project. Use "None" to use the
          current project.
        """
        project = project or self.project.current
        project_info = self.get_project(project)
        ep = self.resolve_endpoint(url)
        if ep and project_info.get('endpoint') != ep:
            raise exc.PilotInvalidProject(
                'URL {} endpoint does not match the given project {} ({} != {}'
                ''.format(url, project, ep, project_info.get('endpoint')))
        purl = urllib.parse.urlparse(url)
        if purl.path.startswith(project_info.get('base_path')):
            return purl.path.replace(project_info.get('base_path'),
                                     '').lstrip('/')
        return purl.path.lstrip('/')

    def get_path(self, path, project=None, relative=True):
        """
        Resolve the absolute Globus endpoint path given a relative project
        path. Raises PilotClientException if relative=False and path is not
        in the project's directory.
        **Parameters**
        ``path`` (*path string*)
          Path to a local resource on this project
        ``project`` (*string*)
          The project to fetch info for. Defaults to current project
        ``relative`` (*bool*)
          If True, prepends the path to the project. If False,
          does not prepend path but ensures it's in the project's directory
        **Examples**
        >>> pc.get_path('foo/bar.txt')
        '/projects/myproject/foo/bar.txt'
        >>> pc.get_path('/projects/goo/moo.txt', project='goo', relative=False)
        '/projects/goo/moo.txt'
        """
        if not isinstance(path, str):
            raise exc.PilotClientException('get_path(): "path" must be a '
                                           'string, not {}'.format(type(path)))
        bdir = self.project.get_info(project)['base_path']
        path = path.lstrip('.')
        if relative is True:
            path = path.lstrip('/')
            if path:
                path = os.path.join(bdir, path.lstrip('/'))
            else:
                path = bdir
        else:
            if bdir not in path:
                log.warning(
                    'Absolute path {} not in project {} path {}'.format(
                        path, project, bdir)
                )
        log.debug('Path: {}'.format(path))
        return path

    def get_globus_http_url(self, path, project=None, relative=True):
        """
        Resolve an http url based on the path.
        **Parameters**
        ``path`` (*path string*)
          Path to a local resource on this project
        ``project`` (*string*)
          The project to fetch info for. Defaults to current project
        ``relative`` (*bool*)
          If True, prepends the path to the project. If False,
          does not prepend path but ensures it's in the project's directory
        **Examples**
        >>> pc.get_globus_http_url('foo/bar.txt')
        'https://5590f6aa-a9f7-4cc3-b130-afc26e2ca0c0.e.globus.org/projects/
         myproject/foo/bar.txt'
        """
        https_host = self.project.get_info().get('https_host')
        info = {'endpoint': self.get_endpoint(project)}
        if not https_host:
            https_host = '{endpoint}.e.globus.org'
        https_host = https_host.format(**info)
        path = self.get_path(path, project, relative)
        parts = ['https', https_host, path, '', '', '']
        return urllib.parse.urlunparse(parts)

    def get_globus_url(self, path, project=None, relative=True):
        """
        Resolve a Globus URL for a resource on a Globus Endpoint
        **Parameters**
        ``path`` (*path string*)
          Path to a local resource on this project
        ``project`` (*string*)
          The project to fetch info for. Defaults to current project
        ``relative`` (*bool*)
          If True, prepends the path to the project. If False,
          does not prepend path but ensures it's in the project's directory
        **Examples**
        >>> pc.get_globus_url('foo/bar.txt')
        'globus://5590f6aa-a9f7-4cc3-b130-afc26e2ca0c0/projects/myproject/
         foo/bar.txt'
        """
        path = self.get_path(path, project, relative)
        parts = ['globus', self.get_endpoint(project), path, '', '', '']
        return urllib.parse.urlunparse(parts)

    def get_globus_app_url(self, path, project=None, relative=True):
        """
        Resolve a URL to the Globus Webapp
        **Parameters**
        ``path`` (*path string*)
          Path to a local resource on this project
        ``project`` (*string*)
          The project to fetch info for. Defaults to current project
        ``relative`` (*bool*)
          If True, prepends the path to the project. If False,
          does not prepend path but ensures it's in the project's directory
        **Examples**
        >>> pc.get_globus_app_url('foo/bar.txt')
        'https://app.globus.org/file-manager?
         origin_id=e55b4eab-6d04-11e5-ba46-22000b92c6ec&
         origin_path=%2FXPCSDATA%2Ftest%2Fnick-testing%2Ffoo%2Fbar.txt'
        """
        path = self.get_path(path, project, relative)
        params = {'origin_id': self.get_endpoint(project), 'origin_path': path}
        return urllib.parse.urlunparse([
            'https', 'app.globus.org', 'file-manager', '',
            urllib.parse.urlencode(params), ''
        ])

    def get_portal_url(self, path=None, project=None):
        """
        Resolve a URL to the Globus Webapp. Relies on the context setting:
          'projects_portal_url'
        Examples include:
          https://myportal.com/myprojects/{{project}}/{{subject}}/'
          https://example.com/{{project}}/foo/bar/{{subject}}/moo/
        **Parameters**
        ``path`` (*path string*)
          Path to a local resource on this project
        ``project`` (*string*)
          The project to fetch info for. Defaults to current project
        **Examples**
        >>> pc.get_portal_url('foo/bar.txt')
        'https://myportal/foo/bar.txt'
        """
        portal_url = self.context.get_value('projects_portal_url')
        if not portal_url:
            return None
        project = project or self.project.current
        subject = ''
        if '{project}' not in portal_url or '{subject}' not in portal_url:
            log.warning('Invalid portal url "{}", the url string requires'
                        '"{{project}} and {{subject}}, for example:'
                        'https://myportal/myprojects/{{project}}/{{subject}}/'
                        .format(portal_url))
        if path:
            sub = self.get_subject_url(path, project=project)
            subject = urllib.parse.quote_plus(urllib.parse.quote_plus(sub))
        portal_url = portal_url.format(project=project, subject=subject)
        # If no subject was given, remove trailing slash
        if portal_url.endswith('//'):
            portal_url = portal_url[0:-1]
        return portal_url

    def get_subject_url(self, path, project=None, relative=True):
        """
        Resolve a subject URL for a given resource, as it should appear as a
        subject in Globus Search. Typically this is the same as a "Globus" URL
        and is equivalent to get_globus_url()
        """
        return self.get_globus_url(path, project, relative)

    def ls(self, path, project=None, relative=True, extended=False):
        """
        Perform a list on the remote endpoint for the given project, and list
        the contents of 'path'. Returns contents in a list.
        **Parameters**
        ``path`` (*path string*)
          Path to a local resource on this project
        ``project`` (*string*)
          The project to fetch info for. Defaults to current project
        ``relative`` (*bool*)
          If True, prepends the path to the project. If False,
          does not prepend path but ensures it's in the project's directory
        ``extended`` (*bool*)
          Return a dict instead for each value, where the key is each item of
          content and the values are all file details for each item.
        **Examples**
        >>> pc.ls('bar')
        ['file1.png', 'foo', 'moo']
        """
        path = self.get_path(path, project, relative)
        endpoint = self.get_endpoint(project)
        project = project or self.project.current
        if not endpoint:
            raise exc.PilotClientException('No Endpoint configured for '
                                           'project {}'.format(project))
        r = self.get_transfer_client().operation_ls(endpoint, path=path)
        if extended:
            return {f['name']: f for f in r['DATA']}
        return [f['name'] for f in r['DATA']]

    def mkdir(self, path, project=None, relative=True):
        """
        Create a directory in the given project
        **Parameters**
        ``path`` (*path string*)
          Path to a local resource on this project
        ``project`` (*string*)
          The project to fetch info for. Defaults to current project
        ``relative`` (*bool*)
          If True, prepends the path to the project. If False,
          does not prepend path but ensures it's in the project's directory
        **Examples**
        >>> pc.mkdir('myfolder')
        """
        rpath = self.get_path(path, project=project, relative=relative)
        tc = self.get_transfer_client()
        tc.operation_mkdir(self.get_endpoint(project), rpath)

    def search(self, project=None, index=None, custom_params=None):
        """
        Perform a search for records in a given project and index, optionally
        with custom_parameters. Returns the raw response from the Globus SDK
        SearchClient.post_search().
        **Parameters**
        ``project`` (*string*)
          The project to fetch info for. Defaults to current project
        ``index`` (*bool*)
          If True, prepends the path to the project. If False,
          does not prepend path but ensures it's in the project's directory
        ``custom_params`` (*dict*)
          Allows setting custom parameters for modifying results returned.
          Standard params include the following:
          search_data = {
            'q': '*',
            'filters': {
                'field_name': 'project_metadata.project-slug',
                'type': 'match_all',
                'values': [project],
            },
            'limit': 100,
            'offset': 0,
          }
          Custom params will override these params (this may result in
          unexpected results from other projects if 'filters' is overrided).
        **Examples**
        Fetch all results in this project:
        >>> pc.list_entries()
        Fetch results in the 'foo' directory:
        >>> pc.list_entries('foo')
        """
        sc = self.get_search_client()
        project = project or self.project.current
        index = index or self.get_index(project=project)
        search_data = {
            'q': '*',
            'filters': {
                'field_name': 'project_metadata.project-slug',
                'type': 'match_all',
                'values': [project],
            },
            'limit': 100,
            'offset': 0,
            'result_format_version': '2017-09-01',
        }
        search_data.update(custom_params or {})
        return sc.post_search(index, search_data).data

    def list_entries(self, path='', project=None, relative=True):
        """Search for files in the given project that match the given path.
        Returns a list of Globus Search GMetaEntries for any matches it finds.
        Paths for files in multi-file collections will return no results,
        use 'pc.get_search_entry' instead to find files in multi-file results.
        Paths matching specific entries will return only that entry.
        **Parameters**
        ``path`` (*path string*)
          Path to a local resource on this project. An empty path will return
          all entries in this project
        ``project`` (*string*)
          The project to fetch info for. Defaults to current project
        ``relative`` (*bool*)
          If True, prepends the path to the project. If False,
          does not prepend path but ensures it's in the project's directory
        **Examples**
        Fetch all results in this project:
        >>> pc.list_entries('')
        Fetch results in the 'foo' directory:
        >>> pc.list_entries('foo')
        """
        project = project or self.project.current
        raw = self.search(project=project)
        log.info('Fetching entry list for project {} path {}'.format(project,
                                                                     path))
        path = self.get_path(path, project=project, relative=relative)
        return [ent for ent in raw['gmeta'] if path in ent.get('subject')]

    def get_full_search_entry(self, path, project=None, relative=True,
                              path_is_sub=False, resolve_collections=True,
                              precise=True):
        """
        Get a search entry for a given resource
        **Parameters**
        ``path`` (*path string*)
          Path to a local resource on this project
        ``project`` (*string*)
          The project to fetch info for. Defaults to current project
        ``relative`` (*bool*)
          If True, prepends the path to the project. If False,
          does not prepend path but ensures it's in the project's directory
        ``path_is_sub`` (*bool*)
          If true, the given path is expected to be the full subject, and is
          passed through without modification
        ``resolve_collections`` (*bool*)
          If the path given points to what might be a multi-file directory
          entry, attempt to resolve the entry.
        ``precise`` (*bool*)
          If the path given points to a location inside a multi-file directory
          only return the record if the location matches a file.
          For example, given an entry containing the files:
            my_dir/foo1.txt, my_dir/foo2.txt, my_dir/foo3.txt
          If precise=True and the path is my_dir/foo4.txt, None will be
          returned. If precise=False and the path is my_dir/foo4.txt, the
          "my_dir" record will still be returned.
        **Examples**
        >>> pc.get_search_entry('foo.txt')
          {'dc': {'creators': [{'creatorName': 'NOAA'}],
            'dates': [{'date': '2019-08-29T19:36:13.064917Z',
                    'dateType': 'Created'}],
             ...
          }
        """
        sc = self.get_search_client()
        project = project or self.project.current
        if path_is_sub:
            subject = path
        else:
            subject = self.get_subject_url(path, project, relative)
        try:
            return sc.get_subject(self.get_index(project), subject,
                                  result_format_version='2017-09-01')
        except globus_sdk.exc.SearchAPIError as sapie:
            if sapie.code == 'NotFound.Generic' and resolve_collections:
                ent = search_discovery.get_sub_in_collection(
                    subject, self.list_entries(), precise=precise
                )
                if ent:
                    return ent

    def get_search_entry(self, path, project=None, relative=True,
                         path_is_sub=False, resolve_collections=True,
                         precise=True):
        entry = self.get_full_search_entry(
            path, project=project, relative=relative, path_is_sub=path_is_sub,
            resolve_collections=resolve_collections, precise=precise
        )
        if entry:
            return entry['content'][0]

    def ingest(self, path, content, group=None, project=None,
               relative=True, index=None, dry_run=False, force=False):
        """
        Ingest content into search. The content is validated against Pilot's
        built-in schema.
        **Parameters**
        ``path`` (*path string*)
          Path to a local resource on this project
        ``content`` (*dict*) Full content dict to ingest
        ``group`` (*uuid string*) Globus group to use for this ingest, if
          different from the group configured for this project. Defaults to
          the group configured for this project
        ``project`` (*string*)
          The project to fetch info for. Defaults to current project
        ``relative`` (*bool*)
          If True, prepends the path to the project. If False,
          does not prepend path but ensures it's in the project's directory
          If the path given points to a location inside a multi-file directory
          only return the record if the location matches a file.
          For example, given an entry containing the files:
            my_dir/foo1.txt, my_dir/foo2.txt, my_dir/foo3.txt
        ``index`` (*uuid string*) Index to ingest to. Defaults to index
          configured for this project
        ``dry_run`` Do not actually ingest, but attempt to construct a gmeta
          entry and validate it.
        **Examples**
        >>> content = pc.gather_metadata('foo.txt', 'bar')
        >>> pc.ingest('bar/foo.txt', content)
        """
        sub = self.get_subject_url(path, project=project, relative=relative)
        if force is False:
            self.validate_subject(sub)
        content = [{'subject': sub, 'content': content}]
        group = group or self.get_group(project)
        gmeta = search.get_gmeta_list(content, default_visible_to=group,
                                      validate=True)
        if dry_run:
            return gmeta
        return self.ingest_gmeta(gmeta, index)

    def ingest_many(self, content_map, group=None, project=None, relative=True,
                    index=None, dry_run=False, force=False):
        """
        Ingest many entries into search, with paths to entries mapped to
        content.
        built-in schema.
        **Parameters**
        ``content_map`` (*dict {short_path: content}*)
          Path to a local resource on this project
        ``group`` (*uuid string*) Globus group to use for this ingest, if
          different from the group configured for this project. Defaults to
          the group configured for this project
        ``project`` (*string*)
          The project to fetch info for. Defaults to current project
        ``relative`` (*bool*)
          If True, prepends the path to the project. If False,
          does not prepend path but ensures it's in the project's directory
          If the path given points to a location inside a multi-file directory
          only return the record if the location matches a file.
          For example, given an entry containing the files:
            my_dir/foo1.txt, my_dir/foo2.txt, my_dir/foo3.txt
        ``index`` (*uuid string*) Index to ingest to. Defaults to index
          configured for this project
        ``dry_run`` Do not actually ingest, but attempt to construct a gmeta
          entry and validate it.
        **Examples**
        >>> content_map = {}
        >>> content_map['bar/foo.txt'] = pc.gather_metadata('foo.txt', 'bar')
        >>> content_map['bar/moo.txt'] = pc.gather_metadata('moo.txt', 'bar')
        >>> pc.ingest_many(content_map)
        """
        content_list = []
        for path, content in content_map.items():
            sub = self.get_subject_url(path, project=project,
                                       relative=relative)
            if force is False:
                self.validate_subject(sub)
            content_list.append({'subject': sub, 'content': content})
        gmeta = search.get_gmeta_list(content_list, group, validate=True)
        if dry_run:
            log.info('{} entry ingest ABORTED due to dry run.'
                     ''.format(len(content_list)))
            return gmeta
        return self.ingest_gmeta(gmeta, index)

    def ingest_gmeta(self, gmeta, index=None):
        """
        Ingest a complete raw gmeta_entry into search.
        **Parameters**
        ``gmeta_entry`` (*dict*)
          A dict matching a Globus Search gmeta entry. An example is here:
          https://globus-sdk-python.readthedocs.io/en/stable/clients/search/#globus_sdk.SearchClient.ingest  # noqa
          {'@version': '2016-11-09',
           'ingest_data': {'@version': '2016-11-09',
                             'gmeta': [{'@version': '2016-11-09',
                                        'content': {},
                                        'id': 'metadata',
                                        'subject': 'foo',
                                        'visible_to': ['public']}]},
           'ingest_type': 'GMetaList'}
        ``index`` (*string*)
          The index to ingest to. Defaults to configured current project index
        **Examples**
        >>> from pilot.search import scrape_metadata, gen_gmeta
        >>> metadata = scrape_metadata('moo.txt')
        >>> subject = pc.get_subject_url('bar/moo.txt')
        >>> group = pc.get_group()
        >>> gmeta = gen_gmeta(subject, [group], metadata)
        >>> pc.ingest_entry(gmeta)
        True
        """
        sc = self.get_search_client()
        index = index or self.get_index()
        result = sc.ingest(index, gmeta)
        pending_states = ['PENDING', 'PROGRESS']
        log.info('Ingesting to {}'.format(index))
        task_status = sc.get_task(result['task_id'])['state']
        while task_status in pending_states:
            log.debug(f'Search task still {task_status}')
            time.sleep(.2)
            task_status = sc.get_task(result['task_id'])['state']
        ingest_task = sc.get_task(result['task_id'])
        if ingest_task['state'] != 'SUCCESS':
            log.error(ingest_task.data)
            raise exc.PilotClientException('Failed to ingest search subject: '
                                           '{}'.format(ingest_task['message']))
        return True

    def gather_metadata(self, dataframe, destination, previous_metadata=None,
                        custom_metadata=None, skip_analysis=False,
                        project=None, foreign_keys=None):
        """Gather metadata on a local file or directory. Returns a new dict
        which combines previous metadata and custom metadata. If skip_analysis
        is True, new analytics won't be attempted and old analytics will be
        carried over.
        **Parameters**
        ``dataframe`` (*path-to-file*)
          Path to a file on the local system
        ``destination`` (*path-string*)
          Path to upload on the remote endpoint, relative to the base path set
          by both the context and project.
        ``previous_metadata`` (*dict*)
          Previous metadata to be combined with the new metadata. This should
          be metadata collected with self.get_search_entry()
        ``custom_metadata`` (*dict*) Custom user provided metadata. Will be
          added to scraped metadata after collection.
        ``skip_analysis`` (*bool*) If true, will attempt to open the file and
          analyze the contents before uploading. This generates extra metadata
          which will be included in search.
        ``project`` (*string*)
          The project to use as the base path. Defaults to current project
        **Examples**
        """
        log.info('Gathering metadata on file {}'.format(dataframe))
        short_path = os.path.join(destination, os.path.basename(dataframe))
        url = self.get_globus_http_url(short_path, project=project)
        new_metadata = search.scrape_metadata(
            dataframe, url, self.profile, self.project.current,
            skip_analysis=skip_analysis
        )
        if foreign_keys:
            base_sub = self.get_subject_url('', project=project)
            existing_paths = [e['subject'].replace(base_sub, '').lstrip('/')
                              for e in self.list_entries('', project=project)]
            new_files = search.get_foreign_keys(new_metadata['files'],
                                                foreign_keys, existing_paths)
            for new_file in new_files:
                d = new_file.get('field_metadata', {}).get('field_definitions')
                for fdef in d:
                    if fdef and fdef.get('reference'):
                        fdef['reference']['resource'] = self.get_subject_url(
                            fdef['reference']['resource'], project=project
                        )
            new_metadata['files'] = new_files
        return search.update_metadata(new_metadata, previous_metadata or {},
                                      custom_metadata or {})

    def update(self, short_path, user_metadata, dry_run=False):
        prev_metadata = self.get_search_entry(short_path)
        new_metadata = search.update_metadata({}, prev_metadata, user_metadata)
        stats = search.gather_metadata_stats(new_metadata, prev_metadata)
        stats['ingest'] = {}
        if stats['metadata_modified'] is False:
            return stats
        stats['ingest'] = self.ingest(short_path, new_metadata,
                                      dry_run=dry_run)
        return stats

    def register(self, dataframe, destination, metadata=None,
                 update=False, dry_run=False, skip_analysis=False,
                 foreign_keys=None):
        """
        Gather metadata on a local search record and register metadata in
        Globus Search. This method assumes either the dataframe already exists
        on a remote endpoint, or that the user will handle uploading the result
        manually. Returns the new metadata for the dataframe.
        This method follows the same behavior as the CLI, raising exceptions
        for several different edge cases, all of which derive from
        pilot.exc.PilotCodeException:
        * pilot.exc.DirectoryDoesNotExist - destination not found
        * pilot.exc.GlobusTransferError - unexpected transfer error
        * pilot.exc.RecordExists - Record exists and update flag was false
        * pilot.exc.DestinationIsRecord -- Can't register record inside another
        * pilot.exc.DryRun - Everything succeeded, but ingest was aborted

        **Parameters**
        ``dataframe`` (*path-to-file*)
          Path to a file on the local system
        ``destination`` (*path-string*)
          Path to upload on the remote endpoint, relative to the base path set
          by both the context and project.
        ``metadata`` (*dict*)
          A dictionary of metadata to include with this upload. Metadata gets
          registered into search to describe this dataframe
        ``update`` (*bool*) Update an existing dataframe. An exception will be
          raised if this is false and a dataframe exists to prevent an existing
          dataframe from being overwritten.
        ``skip_analysis`` (*bool*) If true, will attempt to open the file and
          analyze the contents before uploading. This generates extra metadata
          which will be included in search.
        ``project`` (*string*)
          The project to use as the base path. Defaults to current project
        **Examples**
        """
        dframe = self.get_valid_dataframe(dataframe)
        try:
            self.ls(destination)
        except globus_sdk.exc.TransferAPIError as tapie:
            if tapie.code == 'ClientError.NotFound':
                raise exc.DirectoryDoesNotExist(fmt=[destination]) from None
            else:
                raise exc.GlobusTransferError(tapie.message) from None
        short_path = os.path.join(destination, os.path.basename(dframe))
        subject = self.get_subject_url(short_path)
        # Get a list of all entries to check if the new record already exists
        prev_candidates = self.list_entries()
        prev_entry = search_discovery.get_sub_in_collection(
            subject, prev_candidates, precise=False)
        prev_metadata = {}
        if prev_entry:
            log.debug('Previous entry exists: {}'.format(subject))
            if not update and not dry_run:
                raise exc.RecordExists(prev_entry['content'][0],
                                       fmt=[short_path])
            # If we hit on another subject, use the previous subject. This
            # handles two cases: 1. replacing an existing subject 2. Adding
            # a file to an existing subject, where we should use the top level
            # subject name of the entry.
            subject = prev_entry['subject']
            prev_metadata = prev_entry['content'][0]
        new_metadata = self.gather_metadata(
            dframe, destination, previous_metadata=prev_metadata,
            custom_metadata=metadata or {}, skip_analysis=skip_analysis,
            foreign_keys=foreign_keys
        )
        stats = search.gather_metadata_stats(new_metadata, prev_metadata)
        stats['ingest'] = {}
        if stats['metadata_modified'] is False:
            return stats
        if dry_run:
            return stats
        stats['ingest'] = self.ingest(short_path, new_metadata)
        return stats

    def upload(self, dataframe, destination, metadata=None, globus=True,
               update=False, dry_run=False, skip_analysis=False, project=None,
               foreign_keys=None):
        """
        Register a dataframe in Globus Search then upload it to a relative
        project directory on the configured Globus endpoint.
        Raises all of the exceptions that `register` may raise, plus the
        following (All exceptions derive from pilot.exc.PilotCodeException):
        * pilot.exc.NoLocalEndpointSet - No Endpoint configured in profile
        * pilot.exc.NoDestinationProvided - Bad Destination
        **Parameters**
        ``dataframe`` (*path-to-file*)
          Path to a file on the local system
        ``destination`` (*path-string*)
          Path to upload on the remote endpoint, relative to the base path set
          by both the context and project.
        ``metadata`` (*dict*)
          A dictionary of metadata to include with this upload. Metadata gets
          registered into search to describe this dataframe
        ``globus`` (*bool*) Use globus to upload the dataframe. Requires that
          a local endpoint is set in the current user's profile. If false, an
          http put to the configured Globus endpoint will be done instead.
        ``update`` (*bool*) Update an existing dataframe. An exception will be
          raised if this is false and a dataframe exists to prevent an existing
          dataframe from being overwritten.
        ``skip_analysis`` (*bool*) If true, will attempt to open the file and
          analyze the contents before uploading. This generates extra metadata
          which will be included in search.
        ``project`` (*string*)
          The project to use as the base path. Defaults to current project
        **Examples**
        # With context `base_path` set to '/projects/'
        # With project `base_path` set to 'my-project'
        # Uploads to /projects/my-project/foo.txt
        upload('foo.txt', '/')
        # Uploads to /projects/my-project/my/subdir/foo.txt
        upload('local/dir/foo.txt', 'my/subdir/')
        # Raise an exception before uploading to projects/your-project/foo.txt
        upload('foo.txt',
               '/',
               metadata={'my-meta': 'fooiness'},
               globus=True,
               update=True,
               dry_run=True,
               skip_analysis=True,
               project='your-project'
               )
        """
        dframe = self.get_valid_dataframe(dataframe)
        if globus and not self.profile.load_option('local_endpoint'):
            raise exc.NoLocalEndpointSet()
        if not destination:
            raise exc.NoDestinationProvided(fmt=[self.ls('')])

        stats = self.register(
            dframe, destination, metadata=metadata, update=update,
            dry_run=dry_run, skip_analysis=skip_analysis,
            foreign_keys=foreign_keys
        )
        stats['protocol'] = 'globus' if globus else 'http'
        stats['upload'] = {}
        up = self.upload_globus if globus else self.upload_http
        if not dry_run and stats['files_modified'] is True:
            log.debug('Uploading using {}'.format(up))
            stats['upload'] = up(dframe, destination, project=project)
        return stats

    def upload_http(self, dataframe, destination, project=None):
        """Upload to the configured HTTP endpoint for this context/project.
        Executes a simple upload without any metadata or checking the
        destination for existing files. Overwrites any existing dataframe.
        The project must have a configured http endpoint on petrel
        """
        return_values = []
        for local_path, remote_path in search.get_subdir_paths(dataframe):
            path = self.get_path(os.path.join(destination, remote_path))
            rv = self.get_http_client(project).put(path, filename=local_path)
            return_values.append(rv)
        return return_values

    def get_globus_transfer_paths(self, dataframe, destination, project=None):
        """Returns a list of tuples, with each tuple consisting of a src,
        destination to be used as the 'transfer items' in starting a globus
        transfer."""
        paths = []
        for file_path, remote_short_path in search.get_subdir_paths(dataframe):
            rel_dest = os.path.join(destination, remote_short_path)
            paths.append((file_path, self.get_path(rel_dest, project=project)))
        return paths

    def upload_globus(self, dataframe, destination, project=None,
                      globus_args=None):
        """Upload a dataframe to a project using a Globus Transfer. A local
        endpoint must be configured.
        ** parameters **
        ``dataframe`` (*path-to-file*)
          Path to a file on the local system
        ``destination`` (*path-string*)
          Path to upload on the remote endpoint, relative to the base path set
          by both the context and project.
        ``project`` (*string*)
          The project to use as the base path. Defaults to current project
        ``globus_args`` (*dict*)
          Other arguments to pass to Globus Transfer. Overwrites any defaults.
          See ``transfer_file`` for more info.
          https://globus-sdk-python.readthedocs.io/en/stable/clients/transfer/#globus_sdk.TransferClient.submit_transfer  # noqa
        """
        log.info('Uploading (Globus) {} to {}'.format(dataframe, destination))
        result = self.transfer_files(
            self.profile.load_option('local_endpoint'),
            self.get_endpoint(),
            self.get_globus_transfer_paths(dataframe, destination,
                                           project=project),
            **(globus_args or {})
        )
        tl = transfer_log.TransferLog(self.config)
        dest = os.path.join(destination, os.path.basename(dataframe))
        tl.add_log(result, dest)
        return result

    def transfer_file(self, src_ep, dest_ep, src_path, dest_path,
                      globus_args=None):
        """Low level utility for transferring a single file using Globus.
        Does not account for context/project basepaths.
        ** parameters **
        ``src_ep`` (*uuid*)
          Source Globus Endpoint
        ``destination`` (*uuid*)
          Destination Globus Endpoint
        ``src_path`` (*path-string*)
          Source path to file to-be-transferred
        ``dest_path`` (*path-string*)
          Destination path to transfer file.
        ``globus_args`` (*dict*)
          Globus Transfer options. Defaults include:
          {
            'label': 'MyApp Transfer',
            'notify_on_succeeded': False,
            'sync_level': 'checksum',
            'encrypt_data': True,
          }
          See more options at:
          https://globus-sdk-python.readthedocs.io/en/stable/clients/transfer/#globus_sdk.TransferClient.submit_transfer  # noqa
        """
        return self.transfer_files(src_ep, dest_ep, [(src_path, dest_path)],
                                   globus_args=globus_args)

    def transfer_files(self, src_ep, dest_ep, paths, globus_args=None):
        """Low level utility for transferring a multiple files using Globus.
        Does not account for context/project basepaths. All files are expected
        to originate from the same src_ep, to be transferred to the same
        destination.
        ** parameters **
        ``src_ep`` (*uuid*)
          Source Globus Endpoint
        ``destination`` (*uuid*)
          Destination Globus Endpoint
        ``paths`` (*list of two item tuples*)
          A list of items to transfer. Each item must be a tuple with two
          entries, the first the path to the source file, and the second
          the path of the destination. For example:
          [('/users/foo/bar.txt', '~/bar.txt'), ('a.json', '~/a.json')]
        ``globus_args`` (*dict*)
          Globus Transfer options. Defaults include:
          {
            'label': 'MyApp Transfer',
            'notify_on_succeeded': False,
            'sync_level': 'checksum',
            'encrypt_data': True,
          }
          See more options at:
          https://globus-sdk-python.readthedocs.io/en/stable/clients/transfer/#globus_sdk.TransferClient.submit_transfer  # noqa
        """
        tc = self.get_transfer_client()
        log.debug('Activating {} and {}'.format(src_ep, dest_ep))
        tc.endpoint_autoactivate(src_ep)
        tc.endpoint_autoactivate(dest_ep)
        g_defaults = {
            'label': '{} Transfer'.format(self.context.get_value('app_name')),
            'notify_on_succeeded': False,
            'sync_level': 'checksum',
            'encrypt_data': True,
        }
        g_defaults.update(globus_args or {})
        tdata = globus_sdk.TransferData(tc, src_ep, dest_ep, **g_defaults)
        for src_path, dest_path in paths:
            log.debug('Transferring {} to {}'.format(src_path, dest_path))
            tdata.add_item(src_path, dest_path)
        transfer_result = tc.submit_transfer(tdata)
        log.debug('Submitted Transfer')
        return transfer_result

    def download(self, path, project=None, relative=True, globus=False):
        downloader = self.download_globus if globus else self.download_http
        return downloader(path, project=project, relative=relative)

    def download_parts(self, url, dest=None, project=None, range=None):
        """Download a file in parts over HTTP and yield the number of bytes
        written for each part. Yields a generator for each part."""
        dest = os.path.dirname(dest or '')
        relative_dest = ''
        for dir in dest.split('/'):
            relative_dest = os.path.join(relative_dest, dir)
            if relative_dest and not os.path.exists(relative_dest):
                log.info('Making relative dir: {}'.format(relative_dest))
                os.mkdir(relative_dest)
        http_client = self.get_http_client(project=project or None)
        log.debug(f'Fetching item {url}')
        response = http_client.get(url, range=range)
        file_dest = os.path.join(dest, os.path.basename(url))
        with open(file_dest, 'wb') as fh:
            for part in response.iter_content:
                yield fh.write(part)
        log.debug('Fetch Successful')
        return 0

    def download_http(self, path, dest=None, project=None, relative=True,
                      range=None):
        """
        Download a file to the local system using HTTPS to the local filesystem
        using the 'path' basename. Returns the total number of bytes written
        to the filesystem.
        **Parameters**
        ``path`` (*path string*)
          Path to a local resource on this project
        ``project`` (*string*)
          The project to fetch info for. Defaults to current project
        ``relative`` (*bool*)
          If True, prepends the path to the project. If False,
          does not prepend path but ensures it's in the project's directory
        ``range`` (*string*)
          Specify a byte range to fetch from the file. Range must be in the
          format: "0-10,15-20" where each number is a byte.
        **Examples**
        >>> pc.download('foo.txt')
        >>> pc.download('foo.txt', project='bar', range='0-100')
        >>> pc.download('bar/moo.txt', range='0-100,150-200')
        """
        dest = dest or os.path.basename(path)
        return sum(self.download_parts(self.get_path(path), dest=dest,
                                       project=project, range=range))

    def download_globus(self, path, globus_args=None):
        result = self.transfer_file(
            self.get_endpoint(),
            self.profile.load_option('local_endpoint'),
            path,
            os.path.join(os.getcwd(), os.path.basename(path)),
            **(globus_args or {}))
        return result

    def delete_entry(self, path, entry_id='metadata', full_subject=False,
                     project=None, relative=True, force=False):
        """
        Delete a search entry in Globus Search. If the given path is a partial
        match on a multi-file-entry, the entry is pruned and re-ingested. If
        the delete is partial, full_subject has no effect. If the delete is
        partial and a different entry_id is provided, the entry will be
        re-ingested with the new entry_id.
        **Parameters**
        ``path`` (*path string*)
          Path to a local resource on this project
        ``entry_id`` (*string*)
          Name of the entry to delete.
        ``full_subject`` (*bool*)
          Delete all entries in the subject. This is typically equivalent to
          normal deletes if only one entry exists for the given subject.
        **Examples**
        >>> pc.delete_entry('foo/bar.txt')
        >>> pc.delete_entry('moo.txt', entry_id='special_metadata')
        >>> pc.delete_entry('goo.json' full_subject=True)
        """
        index = self.get_index(project=project)
        entry = self.get_full_search_entry(
            path, project=project, relative=relative, resolve_collections=True,
            precise=True
        )
        if not entry:
            raise exc.RecordDoesNotExist(path)
        sub = entry['subject']
        entry = entry['content'][0]
        full_path = self.get_path(path, project=project, relative=relative)
        search_cli = self.get_search_client()
        if full_subject:
            if force is False:
                self.validate_subject(sub)
            search_cli.delete_subject(index, sub)
        elif not search_discovery.is_top_level(entry, full_path):
            log.info('Pruning {} from multi-file-entry'.format(path))
            new_files = search.prune_files(entry, full_path)
            del_num = len(entry['files']) - len(new_files)
            entry['files'] = new_files
            self.ingest(path, entry)
            return del_num
        else:
            search_cli.delete_entry(index, sub, entry_id=entry_id)

    def delete(self, path, project=None, relative=True, recursive=False):
        """
        Delete a file on the remote endpoint for the given project.
        to the filesystem.
        **Parameters**
        ``path`` (*path string*)
          Path to a local resource on this project
        ``project`` (*string*)
          The project to fetch info for. Defaults to current project
        ``relative`` (*bool*)
          If True, prepends the path to the project. If False,
          does not prepend path but ensures it's in the project's directory
        ``recursive`` (*bool*)
          If the 'path' refers to a folder, delete everything inside it
        **Examples**
        >>> pc.delete('foo.txt')
        >>> pc.delete('bar', recursive=True)
        """
        tc = self.get_transfer_client()
        endpoint = self.get_endpoint(project)
        full_path = self.get_path(path, project=project, relative=relative)
        app_name = self.context.get_value('app_name')
        ddata = globus_sdk.DeleteData(
            tc, endpoint, recursive=recursive, notify_on_succeeded=False,
            label='File Deletion with {}'.format(app_name))
        ddata.add_item(full_path)
        delete_result = tc.submit_delete(ddata)
        log.debug(delete_result)

    def validate_subject(self, subject):
        pdata = self.resolve_project(subject)
        if pdata is None:
            raise exc.InvalidProject('No project associated with subject'
                                     ': {}'.format(subject))
        if pdata['name'] != self.project.current:
            raise exc.SubjectOutsideProject(
                'Subject outside of current project "{}": Subject: {}'
                ''.format(pdata['base_path'], subject))



from copy import deepcopy
from tests.unit.mocks import MOCK_PROJECTS


def test_get_diff(mock_cli):
    other_mock_projects = deepcopy(MOCK_PROJECTS)
    del other_mock_projects['bar-project']

    diff = mock_cli.context.get_diff(other_mock_projects, MOCK_PROJECTS)
    assert set(diff.keys()) == {'added'}
    assert 'bar-project' in diff['added']

    diff = mock_cli.context.get_diff(MOCK_PROJECTS, other_mock_projects)
    assert set(diff.keys()) == {'removed'}
    assert 'bar-project' in diff['removed']

    other_mock_projects['foo-project']['title'] = 'FOOOOOOO'
    diff = mock_cli.context.get_diff(MOCK_PROJECTS, other_mock_projects)
    assert set(diff.keys()) == {'changed', 'removed'}
    assert 'bar-project' in diff['removed']
    from pprint import pprint
    pprint(MOCK_PROJECTS)
    pprint(other_mock_projects)
    assert diff['changed'] == {'foo-project': {'title': 'Foo --> FOOOOOOO'}}



import logging
import pandas as pd
import numpy
import tableschema
import tabulator.exceptions


log = logging.getLogger(__name__)


TSV_LABELS = {
    'name': 'Column Name',
    'type': 'Data Type',
    'format': 'Format',
    'count': 'Number of non-null entries',
    '25': '25th Percentile',
    '50': '50th Percentile',
    '75': '75th Percentile',
    'std': 'Standard Deviation',
    'mean': 'Mean Value',
    'min': 'Minimum Value',
    'max': 'Maximum Value',
    'unique': 'Unique Values',
    'top': 'Top Common',
    'frequency': 'Frequency of Top Common Value',
    'reference': 'Link to resource definition'
}


def analyze_tsv(filename):
    metadata = analyze(pd.read_csv(filename, sep='\t'))
    metadata = add_extended_metadata(filename, metadata)
    return metadata


def analyze_csv(filename):
    metadata = analyze(pd.read_csv(filename))
    metadata = add_extended_metadata(filename, metadata)
    return metadata


def analyze_parquet(filename):
    return analyze(pd.read_parquet(filename))


def analyze_feather(filename):
    return analyze(pd.read_feather(filename))


def analyze_hdf(filename):
    log.debug('Analyzing hdf5!')
    store = pd.HDFStore(filename, 'r')
    analyses = []
    for key in store.keys():
        try:
            if len(key.lstrip('/').split('/')) > 1:
                # Ignore any nested items in the store
                log.debug('Skipping {}'.format(key))
                continue
            log.info('Analyzing HDF key: {}'.format(key))
            dataframe = store.get(key)
            if isinstance(dataframe, pd.Series):
                dataframe = pd.DataFrame({key.lstrip('/'): dataframe})
            analysis = analyze(dataframe)
            analysis['store_key'] = key
            analyses.append(analysis)
        except Exception as e:
            log.exception(e)
            log.error('Error processing key {}'.format(key))
    store.close()
    return analyses


def analyze(pd_dataframe):
    pandas_info = pd_dataframe.describe(include='all')

    column_metadata = []
    for column in pd_dataframe.columns.tolist()[:10]:
        # df_metadata = column.copy()
        # col_name = column['name']
        df_metadata = {'name': column}
        df_metadata.update(get_pandas_field_metadata(pandas_info, column))
        column_metadata.append(df_metadata)

    dataframe_metadata = {
        'name': 'Data Dictionary',
        # df.shape[0] seems to have issues determining rows
        'numrows': len(pd_dataframe.index),
        'numcols': pd_dataframe.shape[1],
        'field_definitions': column_metadata,
        'labels': TSV_LABELS
    }
    return dataframe_metadata


def add_extended_metadata(filename, metadata):
    """Update the given metadata dict with additional tableschema metadata.
    This is only available for files that can be read by tableschema, which are
    only tsvs and csvs. Tableschema doesn't add much, but it could be handy
    if it can detect extended types like locations."""
    metadata['previewbytes'] = get_preview_byte_count(filename)
    try:
        ts_info = tableschema.Schema(tableschema.infer(filename)).descriptor

        new_field_definitions = []
        for m, ts in zip(metadata['field_definitions'], ts_info['fields']):
            m['format'] = ts['format']
            new_field_definitions.append(m)
        metadata['field_definitions'] = new_field_definitions
    except tabulator.exceptions.FormatError:
        pass
    return metadata


def get_preview_byte_count(filename, num_rows=11):
    """Count and return number of bytes for the first 11 rows in the given
    filename. Useful for preview."""
    with open(filename) as fp:
        return sum([len(fp.readline()) for x in range(num_rows)])


def get_pandas_field_metadata(pandas_col_metadata, field_name):
    """
    Fetch information for a given column. The column statistics returned
    will be a bit different depending on if the types in the column are a
    number or a string. 'NAN' values are stripped from statistics and don't
    even show up in output.
    """
    pmeta = pandas_col_metadata.get(field_name)
    # Pandas may return numpy.nan for statistics below, or nothing at all.
    # ALL possibly missing values are treated as NAN values and stripped at
    # the end.
    metadata = {
        'name': field_name,
        'type': 'string' if str(pmeta.dtype) == 'object' else str(pmeta.dtype),
        'count': int(pmeta['count']),
        'top': pmeta.get('top', numpy.nan),

        # string statistics
        'unique': pmeta.get('unique', numpy.nan),
        'frequency': pmeta.get('freq', numpy.nan),

        # numerical statistics
        '25': pmeta.get('25%', numpy.nan),
        '50': pmeta.get('50%', numpy.nan),
        '75': pmeta.get('75%', numpy.nan),
        'mean': pmeta.get('mean', numpy.nan),
        'std': pmeta.get('std', numpy.nan),
        'min': pmeta.get('min', numpy.nan),
        'max': pmeta.get('max', numpy.nan),
    }

    # Remove all NAN values
    cleaned_metadata = {k: v for k, v in metadata.items()
                        if isinstance(v, str) or not numpy.isnan(v)}

    # Pandas has special types for things. Coerce them to be regular
    # ints and floats
    for name in ['25', '50', '75', 'mean', 'std', 'min', 'max']:
        if name in cleaned_metadata:
            cleaned_metadata[name] = float(cleaned_metadata[name])
    for name in ['count', 'unique', 'frequency']:
        if name in cleaned_metadata:
            cleaned_metadata[name] = int(cleaned_metadata[name])
    return cleaned_metadata



import globus_sdk
from unittest.mock import Mock
from click.testing import CliRunner

from pilot.commands.main import cli
from pilot import version


def test_main_no_config(monkeypatch, mock_cli, mock_config):
    mock_cli.config = mock_config
    monkeypatch.setattr(mock_cli, 'is_logged_in', Mock(return_value=False))
    migrate = Mock()
    monkeypatch.setattr(mock_cli.config, 'migrate', migrate)
    runner = CliRunner()
    result = runner.invoke(cli, [])
    assert 'Usage' in result.output
    assert result.exit_code == 0
    assert not migrate.called


def test_main_notify_update(monkeypatch, mock_cli):
    is_cache_stale = Mock(return_value=True)
    monkeypatch.setattr(mock_cli.context, 'is_cache_stale', is_cache_stale)
    update_with_diff = Mock(return_value={'added': 'stuff'})
    monkeypatch.setattr(mock_cli.context, 'update_with_diff', update_with_diff)

    runner = CliRunner()
    runner.invoke(cli, [])
    assert is_cache_stale.called
    assert update_with_diff.called


def test_main_warns_no_project_set(monkeypatch, mock_cli):
    # Ensures we don't call update and try to load tokens
    monkeypatch.setattr(mock_cli.context, 'is_cache_stale',
                        Mock(return_value=False))
    is_set = Mock(return_value=False)
    monkeypatch.setattr(mock_cli.project, 'is_set', is_set)

    runner = CliRunner()
    result = runner.invoke(cli, ['list'])
    assert is_set.called
    assert 'No project set' in result.output


def test_main_error_fetching_projects(monkeypatch, mock_cli,
                                      mock_globus_exception):
    is_cache_stale = Mock(return_value=True)
    monkeypatch.setattr(mock_cli.context, 'is_cache_stale', is_cache_stale)

    monkeypatch.setattr(globus_sdk.exc, 'SearchAPIError',
                        mock_globus_exception)
    err = Mock(side_effect=mock_globus_exception)
    monkeypatch.setattr(mock_cli.context, 'update_with_diff', err)

    runner = CliRunner()
    result = runner.invoke(cli, [])
    assert 'No manifest exists on this index.' in result.output


def test_main_migrate(monkeypatch, mock_cli):
    monkeypatch.setattr(mock_cli.config, 'is_migrated',
                        Mock(return_value=False))
    migrate = Mock()
    monkeypatch.setattr(mock_cli.config, 'migrate', migrate)

    runner = CliRunner()
    result = runner.invoke(cli, [])
    assert migrate.called
    assert 'Old config detected, upgrading... Success!' in result.output


def test_main_migration_failure(monkeypatch, mock_cli):
    monkeypatch.setattr(mock_cli.config, 'is_migrated',
                        Mock(return_value=False))

    migrate = Mock(side_effect=Exception)
    monkeypatch.setattr(mock_cli.config, 'migrate', migrate)

    runner = CliRunner()
    result = runner.invoke(cli, [])
    assert migrate.called
    assert 'Old config detected, upgrading... Failed!' in result.output


def test_main_version(monkeypatch, mock_cli):
    monkeypatch.setattr(mock_cli, 'is_logged_in', Mock(return_value=False))
    runner = CliRunner()
    result = runner.invoke(cli, ['version'])
    assert version.__version__ in result.output



import pytest
import pandas
from io import StringIO
from pilot.analysis import analyze_dataframe
from pilot import exc

from tests.unit.mocks import ANALYSIS_MIXED_FILES

ANALYZABLE_MIMETYPES = [
    'text/csv', 'text/tab-separated-values',
    'application/x-feather', 'application/x-parquet', 'application/x-hdf'
]

EXPECTED_MIXED_ANALYSIS_RESULT = {
    'field_definitions': [
        {'25': 25.5,
         '50': 50.0,
         '75': 74.5,
         'count': 99,
         # Tableschema field, only visible in tsvs, csvs
         # 'format': 'default',
         'max': 99.0,
         'mean': 50.0,
         'min': 1.0,
         'name': 'Numbers',
         'reference': None,
         'std': 28.722813232690143,
         'type': 'float64'},
        {'count': 99,
         # 'format': 'default',
         'frequency': 50,
         'name': 'Title',
         'reference': None,
         'top': 'baz',
         'type': 'string',
         'unique': 3}],
    'labels': {'25': '25th Percentile',
               '50': '50th Percentile',
               '75': '75th Percentile',
               'count': 'Number of non-null entries',
               'format': 'Format',
               'frequency': 'Frequency of Top Common Value',
               'max': 'Maximum Value',
               'mean': 'Mean Value',
               'min': 'Minimum Value',
               'name': 'Column Name',
               'reference': 'Link to resource definition',
               'std': 'Standard Deviation',
               'top': 'Top Common',
               'type': 'Data Type',
               'unique': 'Unique Values'},
    'name': 'Data Dictionary',
    'numcols': 2,
    'numrows': 99,
    # Only viewable in text-based files -- tsvs, csvs
    # 'previewbyptes': 75
}

EXTENDED_ANALYSIS_MIMETYPES = ['text/csv', 'text/tab-separated-values']

ANALYZABLE_MIXED_FILES = [(f, mtype) for f, mtype in ANALYSIS_MIXED_FILES
                          if mtype in ANALYZABLE_MIMETYPES]


@pytest.mark.skip
@pytest.mark.parametrize("filename,mimetype", ANALYZABLE_MIXED_FILES)
def test_analyze_filetypes(filename, mimetype):
    ana = analyze_dataframe(filename, mimetype)
    if mimetype == 'application/x-hdf':
        ana = ana[0]
    assert ana['numcols'] == 2
    assert ana['numrows'] == 99
    row1_keys = set(ana['field_definitions'][0].keys())
    row2_keys = set(ana['field_definitions'][0].keys())
    label_set = set(ana['labels'].keys())
    assert row1_keys.issubset(label_set)
    assert row2_keys.issubset(label_set)
    # The combination of rows 1 and 2 should contain all the labels
    rows_union = row1_keys.union(row2_keys)
    assert not rows_union.isdisjoint(label_set)
    assert ana['labels'] == EXPECTED_MIXED_ANALYSIS_RESULT['labels']
    for result, expected in zip(
            ana['field_definitions'],
            EXPECTED_MIXED_ANALYSIS_RESULT['field_definitions']):
        from pprint import pprint
        pprint(result)
        for field in expected:
            assert result.get(field) == expected.get(field)

    # Check extended stuff
    if mimetype in EXTENDED_ANALYSIS_MIMETYPES:
        assert ana['previewbytes'] == 75
        for field in ana['field_definitions']:
            assert field['format'] == 'default'


@pytest.mark.skip
def test_preview_bytes(mixed_tsv):
    ana = analyze_dataframe(mixed_tsv, 'text/tab-separated-values')
    with open(mixed_tsv) as fp:
        preview_data = StringIO(fp.read(ana['previewbytes']))
    preview_df = pandas.read_csv(preview_data, sep='\t', encoding='utf8')
    normal_df = pandas.read_csv(mixed_tsv, sep='\t')

    assert list(preview_df.columns) == list(normal_df.columns)
    assert preview_df.head(10).to_dict() == normal_df.head(10).to_dict()
    assert preview_df.head(11).to_dict() != normal_df.head(11).to_dict()


def test_analyze_dataframe_with_unknown_mimetype(mixed_tsv):
    assert analyze_dataframe(mixed_tsv, 'completely_unknown_mimetype') == {}


@pytest.mark.skip
def test_analyze_unexpected_error():
    with pytest.raises(exc.AnalysisException):
        analyze_dataframe('does-not-exist', 'text/csv')



import os
import urllib
import json
import logging
import click
from pilot import commands
from pilot.search_parse import (
    parse_result, get_titles, get_common_path, get_relative_paths,
    get_formatted_field_metadata, get_formatted_fields
)
from pilot.search_discovery import get_matching_file

log = logging.getLogger(__name__)


def get_short_path(result):
    pc = commands.get_pilot_client()
    sub = urllib.parse.urlparse(result['subject'])
    return sub.path.replace(pc.get_path(''), '').lstrip('/')


def get_relative_path_from_entries(entry, file_info):
    relative_path = ''
    for path in get_relative_paths(entry):
        if path in file_info['url']:
            relative_path = path
            break
    return relative_path


def get_single_file_info(entry, file_info):
    rp = get_relative_path_from_entries(entry, file_info)
    name = rp or file_info.get('filename', '')
    mimetype = file_info.get('mime_type', '')
    field_metadata = file_info.get('field_metadata')
    info = []
    if not field_metadata:
        return []
    elif isinstance(field_metadata, list):
        for store in field_metadata:
            keystore = store.get('store_key', '')
            info += (
                ['{} ({}) Keystore: {}'.format(name, mimetype, keystore)] +
                get_formatted_field_metadata(store) +
                ['']
            )
    else:
        info = (
            ['{} ({})'.format(name, mimetype)] +
            get_formatted_field_metadata(field_metadata) +
            ['']
        )
    return info


def get_location_info(entry):
    short_path = os.path.basename(get_common_path(entry))
    pc = commands.get_pilot_client()
    return ['Location Information',
            '{:21.20}{}'.format('Subject', pc.get_subject_url(short_path)),
            '{:21.20}{}'.format('Portal', pc.get_portal_url(short_path))]


@click.command(name='list', help='List known records in Globus Search')
@click.argument('path', type=click.Path(), required=False, default='')
@click.option('--json/--no-json', 'output_json', default=False,
              help='Output as JSON.')
@click.option('--limit', type=int, default=100,
              help='Limit returned results to the number provided')
@click.option('--relative/--no-relative', default=True,
              help='Only list results relative to project')
@click.option('--all', 'all_recs', default=False, is_flag=True,
              help='Do not filter on project')
def list_command(path, output_json, limit, relative, all_recs):
    # Should require login if there are publicly visible records
    pc = commands.get_pilot_client()
    project = pc.project.current
    search_params = {'limit': limit}
    if all_recs:
        search_params['filters'], relative = [], False
    search_results = pc.search(project=project, custom_params=search_params)
    log.debug(search_results)
    if output_json:
        click.echo(json.dumps(search_results, indent=4))
        return
    path_sub = pc.get_subject_url(path) if relative else path
    curated_results = [r for r in search_results['gmeta']
                       if path_sub in r['subject']]

    results = 'Showing {}/{} of total results for "{}"'.format(
        len(curated_results), search_results['total'], project)
    items = ['title', 'data', 'dataframe', 'rows', 'columns', 'size']
    titles = get_titles(items) + ['Path']
    fmt = '{:21.20}{:11.10}{:10.9}{:7.6}{:7.6}{:7.6}{}'
    output = []
    for result in curated_results:
        # If this path refers to a result in a different base location, skip
        # it, it isn't part of this project
        if relative and pc.get_path('') not in result['subject']:
            log.debug('Skipping result {}'.format(result['subject']))
            continue

        data = dict(parse_result(result['content'][0], items))
        parsed = [data.get(name) for name in items]
        parsed += [get_short_path(result) if relative else result['subject']]
        parsed = [str(p) for p in parsed]

        output.append(fmt.format(*parsed))
    if not output:
        click.secho('No results to display for "{}"'.format(path or '/'),
                    fg='yellow')
    else:
        output = [results, fmt.format(*titles)] + output
        click.echo('\n'.join(output))

    result_names = [os.path.basename(r['subject']) for r in curated_results]
    path_info = pc.ls(path, extended=True)
    dirs = [name for name, info in path_info.items()
            if info['type'] == 'dir' and name not in result_names]
    if dirs:
        click.echo('\nDirectories:\n\t{}'.format('\n\t'.join(dirs)))
    else:
        click.echo('\nNo Directories in {}'.format(path or '/'))


@click.command(help='Output info about a dataset')
@click.argument('path', type=click.Path())
@click.option('--json/--no-json', 'output_json', default=False,
              help='Output as JSON.')
@click.option('--limit', type=int, default=10,
              help='Limit number of entities displayed')
@click.option('--relative/--no-relative', default=True)
@click.option('--path-is-sub', default=False, is_flag=True)
def describe(path, output_json, limit, relative, path_is_sub):
    pc = commands.get_pilot_client()
    entry = pc.get_search_entry(path, relative=relative,
                                path_is_sub=path_is_sub)
    if not entry:
        click.echo('Unable to find entry')
        return

    if output_json:
        click.echo(json.dumps(entry, indent=4))
        return

    single_file_entry = get_matching_file(pc.get_globus_http_url(path), entry)

    if single_file_entry:
        cols = ['title', 'authors', 'publisher', 'subjects', 'dates',
                'data', 'dataframe', 'rows', 'columns',
                'formats', 'version', 'size', 'description']
        output = '\n'.join(
            get_formatted_fields(entry, cols, limit=limit) +
            [''] +
            get_single_file_info(entry, single_file_entry) +
            ['', ''] +
            get_location_info(entry)
        )
    else:
        cols = ['title', 'authors', 'publisher', 'subjects', 'dates',
                'formats', 'version', 'combined_size', 'description',
                'files']
        output = '\n'.join(
            get_formatted_fields(entry, cols, limit=limit) +
            [''] +
            get_location_info(entry)
        )

    click.echo(output)



import os
import logging
import click
import globus_sdk
import pilot

log = logging.getLogger(__name__)


@click.command(name='delete', help='Delete file and search record')
@click.argument('path', type=click.Path())
@click.option('--entry-id', default='metadata', help=('Delete a specific entry'
              ' within the search subject, or "null" for a null entry id.'))
@click.option('--subject', default=False, is_flag=True, help=('Delete the'
              ' entire subject comprising all of its associated entry ids'))
@click.option('--dry-run', is_flag=True, default=False,
              help="Show report, but don't actually delete entry/file")
@click.option('--data', 'data_only', is_flag=True,
              help='Delete the data but not the search record', default=False)
@click.option('--metadata', 'metadata_only', default=False, is_flag=True,
              help='Delete the search record but not the data')
# @click.option('--recursive', '-r', default=False, is_flag=True,
#               help='Delete the entire contents of a directory')
def delete_command(path, entry_id, subject, dry_run, data_only,
                   metadata_only):
    pc = pilot.commands.get_pilot_client()
    if not pc.is_logged_in():
        click.echo('You are not logged in.')
        return

    is_dir = False
    path = path.rstrip('/')

    if not metadata_only:
        dir_list = pc.ls(os.path.dirname(path), extended=True)
        basename = os.path.basename(path)
        log.debug('Checking if {} in {}'.format(basename, dir_list.keys()))

        if os.path.basename(path) not in dir_list:
            click.secho('No file "{}" exists'.format(path), fg='yellow')
            return
        elif dir_list[basename]['type'] == 'dir':
            is_dir = True

    sub_url = pc.get_subject_url(path)
    if dry_run:
        click.secho('Dry Run (No Delete Performed)')
        if is_dir:
            click.echo('Deleted directory: {}'.format(path))
        if not metadata_only:
            click.echo('Deleted file: {}'.format(path))
        if not data_only:
            click.echo('Deleted search entry: {}'.format(sub_url))
        return

    try:
        if is_dir:
            entry = pc.get_search_entry(path)
            if entry:
                num_files = len(entry.get('files', []))
                deleted = pc.delete_entry(path)
                pc.delete(path, recursive=True)
                if num_files == deleted:
                    click.echo('Removed {} ({} files)'.format(path, num_files))
                else:
                    click.echo('Removed {} ({}/{} files)'.format(
                        path, deleted, num_files))
                return 0
            subdir_listing = pc.ls(path)
            log.debug('Subdir contains {}'.format(subdir_listing))
            if subdir_listing:
                click.secho('Directory is not empty, contains: {}'
                            ''.format(subdir_listing), fg='yellow')
                return 1
            pc.delete(path, recursive=True)
            click.echo('The Directory {} has been removed'.format(path))
            return
        if not metadata_only:
            click.echo(f'Removing {path}... ')
            pc.delete(path)
        if not data_only:
            click.echo(f'Removing search record {path}...')
            pc.delete_entry(path, entry_id=entry_id, full_subject=subject)
        click.secho('{} has been deleted successfully'.format(path),
                    fg='green')
    except globus_sdk.exc.SearchAPIError as se:
        if se.code == 'NotFound.Generic':
            click.secho('\n{} does not exist, or cannot be found at your '
                        'permission level.'.format(path), fg='yellow')
        else:
            click.secho(str(se))



import os
import time
import uuid

BASE_FILE_DIR = os.path.join(os.path.dirname(__file__), 'files')
COMMANDS_FILE_BASE_DIR = os.path.join(BASE_FILE_DIR, 'commands')
CLIENT_FILE_BASE_DIR = os.path.join(BASE_FILE_DIR, 'client')
ANALYSIS_FILE_BASE_DIR = os.path.join(BASE_FILE_DIR, 'analysis')
MIXED_BASE = os.path.join(ANALYSIS_FILE_BASE_DIR, 'types')
BLIND_BASE = os.path.join(ANALYSIS_FILE_BASE_DIR, 'blind_types')
SCHEMA_FILE_BASE_DIR = os.path.join(BASE_FILE_DIR, 'schemas')
MULTI_FILE_DIR = os.path.join(BASE_FILE_DIR, 'multi_file')

MIXED_MIMETYPES = [
    ('csv', 'text/csv'),
    ('tsv', 'text/tab-separated-values'),
    ('pdf', 'application/pdf'),
    # Disabled
    # ('hdf', 'application/x-hdf'),
    ('feather', 'application/x-feather'),
    ('parquet', 'application/x-parquet'),
]

ANALYSIS_MIXED_FILES = [(os.path.join(MIXED_BASE, 'mixed.{}'.format(ext)), mt)
                        for ext, mt in MIXED_MIMETYPES]
ANALYSIS_BLIND_FILES = [(os.path.join(BLIND_BASE, ext), mt)
                        for ext, mt in MIXED_MIMETYPES]

DEFAULT_EXPIRE = int(time.time()) + 60 * 60 * 48

MOCK_INDEX_RECORD = {
    'context': {},
    'projects': {},
    'groups': {},
}

MOCK_CONTEXT = {'test-context': {
    'app_name': 'my app',
    'client_id': 'e6a8d7ab-9087-4f5f-99f1-974446a64a10',
    'manifest_index': '195a17c1-37cc-45b4-9b72-f0f7a154b143',
    'manifest_subject': 'globus://project-manifest.json',
    'projects_portal_url': 'https://myportal/{project}/{subject}/',
    'projects_base_path': '/myroot',
    'projects_cache_timeout': '86400',
    'projects_default_resource_server': 'petrel_https_server',
    'projects_default_search_index': '195a17c1-37cc-45b4-9b72-f0f7a154b143',
    'projects_endpoint': 'foo-project-endpoint',
    'projects_group': '',
    'scopes': ['profile', 'openid',
               'urn:globus:auth:scope:search.api.globus.org:all',
               'urn:globus:auth:scope:transfer.api.globus.org:all']
    }
}

MOCK_PROJECTS = {
    'foo-project': {
        'title': 'Foo',
        'description': 'This is the foo project',
        'endpoint': 'foo-project-endpoint',
        'resource_server': 'foo_https_server',
        'search_index': 'foo-search-index',
        'base_path': '/foo_folder',
        'group': 'foo-group'
    },
    'foo-project-test': {
        'title': 'Foo Test',
        'description': 'This is the foo project',
        'endpoint': 'foo-project-test-endpoint',
        'resource_server': 'foo_https_server',
        'search_index': 'foo-test-search-index',
        'base_path': '/foo_test_folder',
        'group': 'foo-group'
    },
    'bar-project': {
        'description': 'This is the NCI Pilot Project testing project',
        'title': 'NCI Pilot 1 TEST',
        'endpoint': 'bar-project-endpoint',
        'resource_server': 'petrel_https_server',
        'search_index': 'bar-search-index',
        'base_path': '/bar_test_folder',
        'group': 'bar-group'
    },
}


MOCK_TOKEN_SET = {
    'auth.globus.org': {
        'scope': 'openid profile',
        'access_token': '<token>',
        'refresh_token': None,
        'token_type': 'Bearer',
        'expires_at_seconds': DEFAULT_EXPIRE,
        'resource_server': 'auth.globus.org'
    },
    'search.api.globus.org': {
        'scope': 'urn:globus:auth:scope:search.api.globus.org:all',
        'access_token': '<token>',
        'refresh_token': None,
        'token_type': 'Bearer',
        'expires_at_seconds': DEFAULT_EXPIRE,
        'resource_server': 'search.api.globus.org'
    },
    'transfer.api.globus.org': {
        'scope': 'urn:globus:auth:scope:transfer.api.globus.org:all',
        'access_token': '<token>',
        'refresh_token': None,
        'token_type': 'Bearer',
        'expires_at_seconds': DEFAULT_EXPIRE,
        'resource_server': 'transfer.api.globus.org'
    },
    'petrel.http.server': {
        'scope': 'https://auth.globus.org/scopes/'
                 '56ceac29-e98a-440a-a594-b41e7a084b62/all',
        'access_token': '<token>',
        'refresh_token': None,
        'token_type': 'Bearer',
        'expires_at_seconds': DEFAULT_EXPIRE,
        'resource_server': 'petrel.http.server'
    },
    'foo_https_server': {
        'scope': 'https://auth.globus.org/scopes/'
                 '56ceac29-e98a-440a-a594-b41e7a084b62/all',
        'access_token': '<token>',
        'refresh_token': None,
        'token_type': 'Bearer',
        'expires_at_seconds': DEFAULT_EXPIRE,
        'resource_server': 'foo_https_server'
    },

}

MOCK_PROFILE = {
    'name': 'Rosalind Franklin',
    'preferred_username': 'franklinr@globusid.org',
    'organization': 'The French Government Central Laboratory',
    'identity_provider': '41143743-f3c8-4d60-bbdb-eeecaba85bd9',
    'identity_provider_display_name': 'Globus ID',
    'sub': '102e192b-5acb-47ee-80c7-e613d86e7d6a',
    'local_endpoint': str(uuid.uuid4()),
    'local_endpoint_path': '/',
    'local_endpoint_name': 'Fake Auto-Generated Endpoint',
}


class GlobusResponse(object):
    DEFAULT_CLASS_DATA = {}

    def __init__(self, *args, **kwargs):
        self.data = self.DEFAULT_CLASS_DATA

    def __getitem__(self, item):
        return self.data[item]


class GlobusTransferTaskResponse(GlobusResponse):

    def __init__(self, *args, **kwargs):
        super().__init__()
        task_uuid = str(uuid.uuid4())
        self.data = {
            'DATA_TYPE': 'transfer_result',
            'code': 'Accepted',
            'message': 'The transfer has been accepted and a task has been '
                       'created and queued for execution',
            'request_id': 'foobarbaz',
            'resource': '/transfer',
            'submission_id': task_uuid,
            'task_id': task_uuid,
            'task_link': {
                'DATA_TYPE': 'link',
                'href': 'task/{}?format=json'.format(task_uuid),
                'rel': 'related',
                'resource': 'task',
                'title': 'related task'
            }
        }


class MemoryStorage(object):
    def __init__(self):
        super(MemoryStorage, self).__init__()
        self.tokens = {}
        self.data = {}

    def load(self):
        return self.data

    def save(self, data):
        self.data = data

    def write_tokens(self, tokens):
        self.tokens = tokens

    def read_tokens(self):
        return self.tokens

    def clear_tokens(self):
        self.tokens = {}



import os
import stat
import logging
from configobj import ConfigObj
from fair_research_login import ConfigParserTokenStorage

from pilot.version import __version__

log = logging.getLogger(__name__)


class Config:
    def __init__(self, filename):
        self.filename = filename
        self.cfg = None
        cfg = self.load()

        if not cfg:
            cfg['pilot'] = {'version': __version__}
            self.save(cfg)

    def migrate_to_configobj(self):
        cfg = self.load()
        old_cfg = ConfigParserTokenStorage(filename=self.filename)
        cfg['tokens'] = old_cfg.read_tokens()
        cfg['pilot'] = {'version': __version__}
        cfg.write()

    def get_migrator(self):
        """Read the config and fetch the next migration based on the current
        config version."""
        cfg = self.load()
        if cfg and not cfg.get('pilot'):
            return self.migrate_to_configobj

    def migrate(self):
        """Migrate to the newest config version"""
        while not self.is_migrated():
            migrator = self.get_migrator()
            migrator()

    def is_migrated(self):
        return False if self.get_migrator() else True

    def save(self, cfg):
        if self.filename is None:
            return
        cfg.write()
        # Set flags to 600, so only the USER can read and write.
        # This protects tokens from prying eyes on multi-user systems!
        os.chmod(self.filename, stat.S_IREAD | stat.S_IWRITE)

    def load(self):
        if self.cfg:
            return self.cfg
        if self.filename:
            self.cfg = ConfigObj(self.filename)
        else:
            self.cfg = ConfigObj()
        return self.cfg

    def read_tokens(self):
        tokens = self.load().get('tokens', {})
        for tset in tokens:
            tokens[tset]['expires_at_seconds'] = \
                int(tokens[tset]['expires_at_seconds'])
            rt = tokens[tset]['refresh_token']
            tokens[tset]['refresh_token'] = None if rt == 'None' else rt
        return tokens

    def write_tokens(self, tokens):
        cfg = self.load()
        if isinstance(tokens, list):
            tokens = tokens[0]
        cfg['tokens'] = tokens
        cfg.write()

    def clear_tokens(self):
        self.write_tokens({})

    def clear(self):
        cfg = self.load()
        cfg['tokens'] = {}
        cfg.write()


class ConfigSection:
    """A Config Section is a base object which carves out a section for some
    object to use. It's currently used for context, profile, project, etc."""

    SECTION = None

    def __init__(self, config):
        self.config = config
        if not self.SECTION:
            raise NotImplementedError('SECTION must be set on Config Section '
                                      'obj')
        if self.SECTION not in self.config.load():
            cfg = self.config.load()
            cfg[self.SECTION] = {}
            self.config.save(cfg)

    def save_option(self, option, value, section=None):
        cfg = self.config.load()
        section = section or self.SECTION
        if cfg.get(section) is None:
            cfg[section] = {}
        # Configparser takes a literal approach to 'None' and will save it as
        # a string, which can cause issues for things expecting null values.
        # Save as the empty string instead.
        value = '' if value is None else value
        cfg[section][option] = value
        self.config.save(cfg)

    def load_option(self, option, section=None):
        op = self.config.load().get(section or self.SECTION, {}).get(option)
        return op or None



import time
import copy
import logging
import globus_sdk
from pilot import config, exc
from pilot.search import gen_gmeta

log = logging.getLogger(__name__)

# Timeout for when to check for project updates (Default 24 hours)
DEFAULT_PROJECTS_CACHE_TIMEOUT = 60 * 60 * 24


DEFAULT_PILOT_CONTEXT = {
    'client_id': 'e4d82438-00df-4dbd-ab90-b6258933c335',
    'app_name': 'Globus Pilot',
    'manifest_index': None,
    'manifest_subject': 'globus://project-manifest.json',
    'scopes': [
        'profile',
        'openid',
        'urn:globus:auth:scope:search.api.globus.org:all',
        'urn:globus:auth:scope:transfer.api.globus.org:all',
        'https://auth.globus.org/scopes/'
        '56ceac29-e98a-440a-a594-b41e7a084b62/all',
    ],
    'projects_cache_timeout': DEFAULT_PROJECTS_CACHE_TIMEOUT,
    'projects_endpoint': '',
    'projects_base_path': '',
    'projects_group': '',
    'projects_default_search_index': None,
    'projects_default_resource_server': 'petrel_https_server',
}


class Context(config.ConfigSection):

    SECTION = 'context'
    DEFAULT_CONTEXT = 'candle-pilot1'

    def __init__(self, client, *args, index_uuid=None, **kwargs):
        self.client = client
        self._current = None
        super().__init__(*args, **kwargs)
        if index_uuid:
            display_name = self.add_context_by_uuid(index_uuid)
            self.current = display_name
            self.update()

    @property
    def current(self):
        curr = self.load_option('current') or self._current
        if curr is None:
            raise exc.PilotContextException('No current context configured')
        return curr

    @current.setter
    def current(self, value):
        contexts = self.load_option('contexts') or {}
        ctx_names = list(contexts.keys())
        if ctx_names and value not in ctx_names and value is not None:
            raise ValueError(f'Context must be one of: {", ".join(contexts)}')
        self.save_option('current', value)

    def load_all(self):
        return self.config.load().get('contexts', {})

    def add_context_by_uuid(self, index_uuid):
        ctx = copy.deepcopy(DEFAULT_PILOT_CONTEXT)
        index_info = self.get_index(index_uuid)
        display_name = index_info['display_name']
        ctx['manifest_index'] = index_uuid
        self.add_context(display_name, ctx)
        return display_name

    def add_context(self, name, context):
        ctx = copy.deepcopy(DEFAULT_PILOT_CONTEXT)
        ctx.update(context)
        self.save_option(name, ctx, section='contexts')

    def get_context(self, context=None):
        return self.load_option(context or self.current, section='contexts')

    def update_context(self, new_context_info, context=None):
        nci_ks = set(new_context_info.keys())
        key_diff = nci_ks.difference(set(DEFAULT_PILOT_CONTEXT.keys()))
        if key_diff:
            raise exc.PilotContextException(f'Invalid context keys set: '
                                            f'{key_diff}')
        ctx_name = context or self.current
        ctx = self.get_context(ctx_name)
        ctx.update(new_context_info)
        self.save_option(ctx_name, ctx, section='contexts')

    def set_context(self, context):
        if self.current == context:
            return
        self.current = context
        self.update()

    def get_value(self, field, context=None):
        return self.get_context(context).get(field)

    def get_index(self, index_uuid):
        return self.get_search_client().get_index(index_uuid).data

    def get_search_client(self):
        try:
            sc = self.client.get_search_client()
        except Exception:
            log.debug('Failed to get authenticated search client, '
                      'fetching unauthenticated one instead.', exc_info=True)
            sc = globus_sdk.SearchClient()
        return sc

    def update(self, index=None, dry_run=False, update_groups_cache=True):
        """Update the local list of projects and groups."""
        self.reset_cache_timer()
        sub = self.get_value('manifest_subject')
        index = index or self.get_value('manifest_index')
        log.debug('Fetching manifest {} from index {}'.format(sub, index))
        try:
            sc = self.get_search_client()
            result = sc.get_subject(index, sub,
                                    result_format_version='2019-08-27')
            manifest = result.data['entries'][0]['content']
        except globus_sdk.exc.SearchAPIError as sapie:
            if sapie.code == 'NotFound.Generic':
                self.client.project.purge()
                raise exc.NoManifestException(
                    'No existing context data found for {}.'
                    ''.format(self.get_value('manifest_subject')))
            else:
                log.debug('Encountered error updating context',
                          exc_info=True)
                raise exc.PilotClientException('Unexpected Error {}'.format(
                    str(sapie)))
        if dry_run is False:
            log.debug('Writing fresh context to config.')
            cfg = self.config.load()
            index_name = sc.get_index(index).data['display_name']
            context = manifest.get('context')
            if context:
                cfg['contexts'][index_name] = context
            cfg['projects'] = manifest.get('projects', {})
            cfg['groups'] = manifest.get('groups', {})
            cfg.write()
        return manifest

    def update_with_diff(self, index=None, dry_run=False,
                         update_groups_cache=True):
        new = self.update(index=index, dry_run=dry_run,
                          update_groups_cache=update_groups_cache)
        projects = self.client.project.load_all()
        groups = self.client.project.load_groups()
        return {
            'context': self.get_diff(self.get_context(),
                                     new.get('context', {})),
            'projects': self.get_diff(projects, new.get('projects', {})),
            'groups': self.get_diff(groups, new.get('groups', {}))
        }

    def get_diff(self, old, new):
        """Fetch the differences between two dictionaries. Dicts can be one
        or two levels deep, for example:
        old: {'foo': 'bar'}, new: {'foo': 'moo'}
        OR
        old: {'foo': {'bar': 'baz'}, new: {'foo': {'bar': 'moo'}
        """
        oldk, newk = set(old.keys()), set(new.keys())
        diff = dict()
        diff['removed'] = {k: old[k] for k in oldk - newk}
        diff['added'] = {k: new[k] for k in newk - oldk}
        diff['changed'] = {}
        for k in oldk.intersection(newk):
            if old[k] != new[k]:
                if isinstance(old[k], str):
                    diff['changed'][k] = '"{}" changed to "{}"'.format(old[k],
                                                                       new[k])
                else:
                    changed = [pk for pk in set(old[k]).union(set(new[k]))
                               if old[k].get(pk) != new[k].get(pk)]
                    changed_str = [f'{old[k].get(c)} --> {new[k].get(c)}'
                                   for c in changed]
                    diff['changed'][k] = dict(zip(changed, changed_str))
        return {k: v for k, v in diff.items() if v}

    def reset_cache_timer(self):
        self.save_option('last_update', int(time.time()))

    def is_cache_stale(self):
        if not self.get_context():
            return True
        last_updated = self.load_option('last_update')
        if last_updated:
            cfg_timeout = int(
                self.get_value('projects_cache_timeout'))
            if time.time() < int(last_updated) + cfg_timeout:
                return False
        return True

    def push(self, context=None):
        context_info = self.get_context(context or self.current)
        manifest = {
            'projects': dict(self.client.project.load_all()),
            'groups': dict(self.client.project.load_groups()),
            'context': dict(context_info)
        }
        gmeta = gen_gmeta(context_info['manifest_subject'], ['public'],
                          manifest, validate=False)
        self.client.ingest_gmeta(gmeta, index=context_info['manifest_index'])

    def fetch_subgroups(self, group=None):
        nc = self.client.get_nexus_client()
        if nc is not None:
            ctx_group = self.get_value('projects_group')
            resp = nc.get_subgroups(group or ctx_group)
            return resp.data.json().get('children')
        return []



import logging
import globus_sdk
import pilot.commands
import pilot.exc

log = logging.getLogger(__name__)


def test_local_endpoint():
    pc = pilot.commands.get_pilot_client()
    tc = pc.get_transfer_client()
    try:
        tc.operation_ls(pc.profile.load_option('local_endpoint'))
    except globus_sdk.exc.TransferAPIError as tapie:
        log.exception(tapie)
        raise pilot.exc.LocalEndpointUnresponsive(fmt=[tapie.message])



import pytest
from pilot.analysis.mimetypes import (
    detect_type, detect_parquet, detect_feather, detect_hdf,
    detect_delimiter_separated_values, get_text_or_binary
)

from tests.unit.mocks import ANALYSIS_MIXED_FILES, ANALYSIS_BLIND_FILES

# Analyzers that test for specific mimetypes. Don't bother testing the general
# detectors.
SPECIFIC_DETECTORS = [
    (detect_parquet, ['application/x-parquet']),
    (detect_feather, ['application/x-feather']),
    (detect_hdf, ['application/x-hdf']),
    (detect_delimiter_separated_values, ['text/csv',
                                         'text/tab-separated-values']),
]
ALL_MIMETYPE_FILES = ANALYSIS_MIXED_FILES + ANALYSIS_BLIND_FILES


@pytest.mark.parametrize("filename,mimetype", ANALYSIS_MIXED_FILES)
def test_postulate_type(filename, mimetype):
    assert detect_type(filename) == mimetype


@pytest.mark.parametrize("filename,mimetype", ANALYSIS_BLIND_FILES)
def test_postulate_type_without_extensions(filename, mimetype):
    skip = []
    if mimetype not in skip:
        assert detect_type(filename) == mimetype


@pytest.mark.parametrize("function,mimetypes", SPECIFIC_DETECTORS)
def test_mimetype_analyzers_against_all_types(function, mimetypes):
    all_types = ANALYSIS_MIXED_FILES + ANALYSIS_BLIND_FILES
    for filename, mimetype in all_types:
        try:
            detected_type = function(filename)
            if detected_type:
                assert detected_type in mimetypes
                assert detected_type == mimetype
            else:
                assert detected_type not in mimetypes
        except AssertionError:
            raise
        except Exception:
            # It's ok if the individual detectors raise errors. Raising errors
            # is another way to say the type could not be determined, and
            # individual detectors should never be run outside 'postulate_type'
            pass


def test_text_or_binary():
    for filename, mimetype in ALL_MIMETYPE_FILES:
        print(filename)
        if mimetype.startswith('text'):
            print((filename, mimetype))
            assert get_text_or_binary(filename) == 'text/plain'
        else:
            assert get_text_or_binary(filename) == 'application/octet-stream'
        print('{} Success'.format(mimetype))



import os
import re
import logging
import globus_sdk
import click

import pilot
import pilot.exc
from pilot.commands import input_validation, endpoint_utils


log = logging.getLogger(__name__)


@click.command(help='Login with Globus')
@click.option('--refresh-tokens/--no-refresh-tokens', default=True,
              help='Request a refresh token to login indefinitely')
@click.option('--force/--no-force', default=False,
              help='Do a fresh login, ignoring any existing credentials')
@click.option('--local-server/--no-local-server', default=True,
              help='Start a local TCP server to handle the auth code')
@click.option('--browser/--no-browser', default=True,
              help='Automatically open the browser to login')
def login(refresh_tokens, force, local_server, browser):
    pc = pilot.commands.get_pilot_client()
    is_logged_in = pc.is_logged_in()
    if is_logged_in and not force:
        click.echo('You are already logged in.')
        return
    elif is_logged_in and force:
        pc.logout()

    prev_info = pc.profile.load_user_info()
    scopes = pc.context.get_value('scopes') or pc.DEFAULT_SCOPES
    pc.login(refresh_tokens=refresh_tokens,
             no_local_server=not local_server,
             no_browser=not browser,
             force=force,
             requested_scopes=scopes)
    if not pc.project.load_all():
        log.debug('NO project info saved, updating...')
        pc.context.update_with_diff()
    click.secho('You have been logged in.', fg='green')

    local_ep = (pc.profile.load_option('local_endpoint') or
                globus_sdk.LocalGlobusConnectPersonal().endpoint_id)
    local_path = pc.profile.load_option('local_endpoint_path')
    log.debug('Local Endpoint set to {}:{}'.format(local_ep, local_path))
    tc = pc.get_transfer_client()
    try:
        if local_ep:
            name = tc.get_endpoint(local_ep).data['display_name']
            pc.profile.save_option('local_endpoint', local_ep)
            pc.profile.save_option('local_endpoint_path', local_path)
            pc.profile.save_option('local_endpoint_name', name)
            endpoint_utils.test_local_endpoint()
    except pilot.exc.LocalEndpointUnresponsive as leu:
        log.debug('Endpoint UUID: {}, local path: {}'
                  .format(local_ep, local_path))
        click.secho(str(leu), fg='yellow')
    if prev_info != pc.profile.load_user_info():
        pitems = [('Name:', pc.profile.name),
                  ('Organization:', pc.profile.organization),
                  ('Local Endpoint:', pc.profile.load_option(
                      'local_endpoint_name'))]
        pstr = '\n'.join(['{:16}{}'.format(t, v) for t, v in pitems])
        report = (
            'Your personal info has been saved as: \n{}\n\n'
            'You can update these with "pilot profile -i"'.format(pstr)
        )
        click.secho(report, fg='blue')


@click.command(help='Revoke local tokens')
@click.option('--purge', default=False, is_flag=True,
              help='Clear all transfer logs and user info')
def logout(purge):
    pc = pilot.commands.get_pilot_client()
    if pc.is_logged_in():
        pc.logout()
        click.secho('You have been logged out.', fg='green')
    else:
        click.echo('No user logged in, no tokens to clear.')
    if purge and os.path.exists(pc.config_file):
        os.unlink(pc.config_file)
        click.secho('All local user info and logs have been deleted.',
                    fg='green')


@click.command(name='profile', help='Output Globus Identity used to login')
@click.option('-i', '--interactive', default=False, is_flag=True,
              help='Interactively set all profile options')
@click.option('--local-endpoint', help='Set local endpoint (ep:/my/path)')
def profile_command(interactive, local_endpoint):
    pc = pilot.commands.get_pilot_client()
    profile_queries = {
        'name': {
            'prompt': 'Name', 'default': pc.profile.name,
            'help': 'When you upload files, your name will be used as the '
                    'author',
            'validation': [],
        },
        'organization': {
            'prompt': 'Organization', 'default': pc.profile.organization,
            'help': 'When you upload files, your organization will be listed',
            'validation': [],
        },
    }
    if local_endpoint:
        pattern = r'(?P<endpoint>[\w-]{36})(:(?P<path>[\w/~]+))?'
        match = re.match(pattern, local_endpoint)
        if not match:
            click.secho('Provide your local endpoint in the notation: '
                        'myendpoint:/my/path', fg='red')
            return
        tc = pc.get_transfer_client()
        matchd = match.groupdict()
        try:
            tc.operation_ls(matchd['endpoint'], path=matchd.get('path'))
            name = tc.get_endpoint(matchd['endpoint']).data['display_name']
            pc.profile.save_option('local_endpoint', matchd['endpoint'])
            pc.profile.save_option('local_endpoint_path', matchd['path'])
            pc.profile.save_option('local_endpoint_name', name)
            click.secho('Your local endpoint has been set!', fg='green')
        except globus_sdk.exc.TransferAPIError as tapie:
            click.secho('Unable to access endpoint, please choose a path where'
                        ' you have read/write access: {}'
                        ''.format(tapie.message), fg='red')

    if interactive:
        od = ['name', 'organization']
        iv = input_validation.InputValidator(queries=profile_queries, order=od)
        info = pc.profile.load_user_info()
        new_profile = iv.ask_all()
        info.update(new_profile)
        pc.profile.save_user_info(info)
        click.secho('Your information has been updated', fg='green')
        return
    info = pc.profile.load_user_info()
    pitems = [('Name:', pc.profile.name),
              ('Organization:', pc.profile.organization),
              ('Identity: ', pc.profile.load_option('preferred_username')),
              ('Local Endpoint:', info.get('local_endpoint_name')),
              ('Local Path:', info.get('local_endpoint_path')),
              ('Endpoint UUID:', info.get('local_endpoint')),
              ('Config File:', pc.config_file),
              ]
    pstr = '\n'.join(['{:16}{}'.format(t, v) for t, v in pitems])
    click.echo('Your Profile: \n{}\n'.format(pstr))



import os
from pilot.search_discovery import get_sub_in_collection, is_top_level
from tests.unit.mocks import CLIENT_FILE_BASE_DIR

MULTI_FILE_METADATA = os.path.join(CLIENT_FILE_BASE_DIR,
                                   'multi_file_entry.json')


def test_get_sub_in_collection(mock_multi_file_result):
    gmeta = mock_multi_file_result['gmeta']
    content = gmeta[0]
    mf_sub = gmeta[0]['subject']
    mf_tsv = os.path.join(mf_sub, 'folder/folder2/tsv2.tsv')
    no_exist = os.path.join(mf_sub, 'does_not_exist.csv')

    assert get_sub_in_collection(mf_sub, gmeta, precise=True) == content
    assert get_sub_in_collection(mf_tsv, gmeta, precise=True) == content
    assert get_sub_in_collection(no_exist, gmeta, precise=True) is None
    assert get_sub_in_collection(mf_tsv, gmeta, precise=False) == content


def test_get_sub_in_collection_none_on_non_single(mock_multi_file_result,
                                                  mock_cli):
    gmeta = mock_multi_file_result['gmeta']
    content1, sub1 = gmeta[0], gmeta[0]['subject']
    sub2 = mock_cli.get_subject_url('another_sub')
    url2 = mock_cli.get_globus_http_url('another_sub')
    ent2 = {'content': [{'files': [{'url': url2}]}], 'subject':  sub2}
    content2 = ent2

    gmeta.append(ent2)

    assert get_sub_in_collection(sub1, gmeta, precise=True) == content1
    assert get_sub_in_collection(sub2, gmeta, precise=True) == content2


def test_get_sub_in_collection_similar_subs(mock_multi_file_result, mock_cli):
    """
    Test for regression where similar subjects would match one another. For
    example:
    globus://foo-project-endpoint/foo_folder/foo
    globus://foo-project-endpoint/foo_folder/foofoo

    Searching for 'globus://foo-project-endpoint/foo_folder/foo' would match
    both above, resulting in the subject not being returned at all.
    """
    subs = ['foo', 'foofoo', 'bar']
    entries = [
        {
            'subject': mock_cli.get_subject_url(s),
            'content': [{'files': [{'url': mock_cli.get_globus_http_url(s)}]}]
        } for s in subs
    ]
    foofoo = mock_cli.get_subject_url('foofoo')
    foo = mock_cli.get_subject_url('foo')
    assert get_sub_in_collection(foofoo, entries, precise=True) is not None
    assert get_sub_in_collection(foofoo, entries, precise=False) is not None
    assert get_sub_in_collection(foo, entries, precise=True) is not None
    assert get_sub_in_collection(foo, entries, precise=False) is not None


def test_get_sub_in_collection_similar_file(mock_cli):
    files = ['foofoo']
    entries = [
        {
            'subject': mock_cli.get_subject_url('ent'),
            'content': [
                {'files': [
                    {'url': mock_cli.get_globus_http_url(f)} for f in files
                ]
                }
            ]
        }
    ]
    foo = mock_cli.get_subject_url('ent/foo')
    # File does not exist, but still produces match
    assert get_sub_in_collection(foo, entries, precise=False) is not None
    # Precise match required, returns None
    assert get_sub_in_collection(foo, entries, precise=True) is None


def test_get_sub_with_dirs(mock_multi_file_result, mock_cli):
    gmeta = mock_multi_file_result['gmeta']
    gmeta[0]['subject'] = mock_cli.get_subject_url('dir/file1.txt')
    sub2 = mock_cli.get_subject_url('dir/file2.txt')
    assert get_sub_in_collection(sub2, gmeta, precise=False) is None


def test_is_top_level(mock_multi_file_result):
    entry = mock_multi_file_result['gmeta'][0]['content'][0]
    assert is_top_level(entry, '/foo_folder/multi_file/') is True
    assert is_top_level(entry, '/foo_folder/multi_file') is True
    assert is_top_level(entry,
                        '/foo_folder/multi_file/text_metadata.txt') is False
    assert is_top_level(entry, '/foo_folder/multi_file/folder/') is False
    assert is_top_level(entry,
                        '/foo_folder/multi_file/folder/folder2') is False
    assert is_top_level(entry, '/foo_folder/multi_file/does_not_exst') is False



from pilot import config


class Profile(config.ConfigSection):

    SECTION = 'profile'

    def load_user_info(self):
        return self.config.load()['profile']

    def save_user_info(self, user_info):
        old_info = self.load_user_info()
        if user_info['sub'] != old_info.get('sub'):
            cfg = self.config.load()
            cfg['profile'] = user_info
            # Also clear transfer logs
            self.config.save(cfg)

    @property
    def name(self):
        return self.load_option('name') or ''

    @name.setter
    def name(self, value):
        self.save_option('name', value)

    @property
    def organization(self):
        return self.load_option('organization') or ''

    @organization.setter
    def organization(self, value):
        self.save_option('organization', value)



import copy
import uuid
from click.testing import CliRunner
from tests.unit.mocks import MOCK_INDEX_RECORD, GlobusResponse
from pilot.commands.project.index import set_index, info


def test_set_index_with_prev_fetched_index(monkeypatch, mock_cli, mock_config,
                                           mock_search_client,
                                           mock_search_result):
    # Result when pilot context calls get_subject on the manifest record
    manifest = copy.deepcopy(mock_search_result)
    # This should be removed when we upgrade to the newer style search results
    manifest.pop('content')
    manifest['entries'] = [{'content': MOCK_INDEX_RECORD}]

    sub_resp = GlobusResponse()
    sub_resp.data = manifest
    mock_search_client.get_subject.return_value = sub_resp

    # Result when pilot fetches the general index_data
    index_resp = GlobusResponse()
    index_resp.data = {'display_name': 'foo'}
    mock_search_client.get_index.return_value = index_resp

    runner = CliRunner()
    result = runner.invoke(set_index, ['test-context'])
    assert result.exit_code == 0
    assert mock_search_client.get_subject.called
    assert mock_search_client.get_subject


def test_set_index_with_non_uuid(mock_cli):
    runner = CliRunner()
    result = runner.invoke(set_index, ['foo'])
    assert result.exit_code != 0
    assert 'must be a UUID' in result.output


def test_set_index_with_new_index(monkeypatch, mock_cli, mock_config,
                                  mock_search_client,
                                  mock_search_result):
    # Result when pilot context calls get_subject on the manifest record
    manifest = copy.deepcopy(mock_search_result)
    manifest.pop('content')
    manifest['entries'] = [{'content': MOCK_INDEX_RECORD}]
    sub_resp = GlobusResponse()
    sub_resp.data = manifest
    mock_search_client.get_subject.return_value = sub_resp

    # Result when pilot fetches the general index_data
    index_resp = GlobusResponse()
    index_resp.data = {'display_name': 'foo'}
    mock_search_client.get_index.return_value = index_resp

    # make a fake index
    fake_index = str(uuid.uuid4())

    runner = CliRunner()
    result = runner.invoke(set_index, [fake_index])
    assert result.exit_code == 0


def test_get_info(mock_cli):
    runner = CliRunner()
    result = runner.invoke(info, ['test-context'])
    assert result.exit_code == 0



import os
from setuptools import setup, find_packages

# single source of truth for package version
version_ns = {}
with open(os.path.join('pilot', 'version.py')) as f:
    exec(f.read(), version_ns)

with open('README.rst') as f:
    long_description = f.read()

install_requires = []
with open('requirements.txt') as reqs:
    for line in reqs.readlines():
        req = line.strip()
        if not req or req.startswith('#'):
            continue
        install_requires.append(req)

setup(
    name='globus-pilot',
    description='A CLI tool for cataloging pilot 1 data',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/globusonline/pilot1-tools/tree/master',
    maintainer='globus',
    maintainer_email='support@globus.org',
    version=version_ns['__version__'],
    packages=find_packages(),
    package_data={
      '': ['*.json'],
    },
    requires=[],
    entry_points='''
    [console_scripts]
    pilot=pilot.commands.main:cli
''',
    install_requires=install_requires,
    dependency_links=[],
    classifiers=[
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ]
)



import logging
import mimetypes
import puremagic
try:
    import pandas as pd
except ImportError:
    pd = None

log = logging.getLogger(__name__)


def detect_type(url, functions=None):
    """This function mimics mimetypes.guess_type, but attempts to open the
    file and read data to determine what the type is."""
    for guesser, name in functions or MIMETYPE_DETECTORS:
        try:
            guess = guesser(url)
            if guess and any(guess):
                log.debug('Determined mimetype to be {}'.format(name))
                return guess
        except Exception:
            log.debug('Attempt to guess mimetype with {} failed'.format(name))


def general_mimetype(url):
    mt, _ = mimetypes.guess_type(url, strict=True)
    return mt


def puremagic_mimetype(url):
    return puremagic.magic_file(url)[0].mime_type


def detect_parquet(url):
    if not pd.read_parquet(url).empty:
        return 'application/x-parquet'


def detect_feather(url):
    if not pd.read_feather(url).empty:
        return 'application/x-feather'


def detect_hdf(url):
    store = pd.HDFStore(url, 'r')
    keys = store.keys()
    store.close()
    if len(keys) > 0:
        return 'application/x-hdf'


def detect_delimiter_separated_values(filename):
    """Attempts to check for csv or tsv mimetypes"""
    df = pd.read_csv(filename)
    if len(df.columns) > 1:
        return 'text/csv'
    if len(df.columns) == 1:
        if df.columns[0].count('\t') >= 1:
            return 'text/tab-separated-values'


def get_text_or_binary(filename):
    """Read the first 1024 and attempt to decode it in utf-8. If this succeeds,
    the file is determined to be text. If not, its binary."""
    with open(filename, 'rb') as f:
        chunk = f.read(1024)
    try:
        chunk.decode('utf-8')
        return 'text/plain'
    except UnicodeDecodeError:
        return 'application/octet-stream'


MIMETYPE_DETECTORS = [
    (general_mimetype, []),
    (puremagic_mimetype, []),
    (detect_parquet, ['application/x-parquet']),
    (detect_feather, ['application/x-feather']),
    (detect_hdf, ['application/x-hdf']),
    (detect_delimiter_separated_values, ['text/csv',
                                         'text/tab-separated-values']),
    (get_text_or_binary, ['text/plain', 'application/octet-stream']),
]



import sys
import os
import uuid
import logging
import click
import globus_sdk

from pilot import exc, commands

log = logging.getLogger(__name__)


class InputValidator:

    def __init__(self, queries=None, order=None):

        self.queries = queries or {}
        self.order = order or sorted(self.queries.keys())
        if not set(order) == set(queries.keys()):
            raise ValueError('Order given must match number of queries: \n'
                             f'Queries: {set(queries)}\nOrder: {set(order)}')
        self.answers = {}

    def ask(self, query, info):
        default = info.get('default')
        if callable(default):
            default = default(self, query)
        user_input = input(f'{info["prompt"]} ({default})> ') or default
        if user_input == 'help':
            click.echo(info['help'])
            return self.ask(query, info)
        if user_input == 'q':
            click.secho('Exiting...', fg='red')
            sys.exit(-1)
        try:
            for validate in info.get('validation', []):
                validate(self, user_input)
            self.answers[query] = user_input
            return user_input
        except exc.PilotValidator as iv:
            click.secho(iv.message, fg='yellow')
        except Exception as e:
            click.secho(e, fg='red')
        return self.ask(query, info)

    def ask_all(self):
        click.secho(f'{len(self.queries)} items to set. Type "help" for more '
                    f'information or "q" to quit.', fg='blue')
        for query in self.order:
            self.ask(query, self.queries[query])
        click.echo('Summary:')
        for subject, answer in self.answers.items():
            title = self.queries[subject].get('title', subject)
            click.echo('{:20.19}{}'.format(title, answer))
        if input('Continue with these values? (y/n)') != 'y':
            for k, v in self.answers.items():
                self.queries[k]['default'] = v
            return self.ask_all()
        return self.answers


def validate_no_spaces(v, string):
    if ' ' in string:
        raise exc.PilotValidator(f'"{string}" cannot contain spaces')


def validate_absolute_path(v, string):
    if not os.path.isabs(string):
        raise exc.PilotValidator('Path must be absolute')


def validate_no_tilde(v, string):
    if '~' in string:
        raise exc.PilotValidator('No "~" allowed in path')


def validate_project_title_unique(v, title):
    titles = v.queries['title']['current']
    if title in titles:
        raise exc.PilotValidator('Title must be unique from '
                                 f'{", ".join(titles)}')


def validate_project_slug_unique(v, slug):
    pc = commands.get_pilot_client()
    slugs = pc.project.load_all().keys()
    if slug in slugs:
        raise exc.PilotValidator(f'Slug must be unique from {", ".join(slug)}')


def validate_slug_to_path_unique(v, slug):
    pc = commands.get_pilot_client()
    tc = pc.get_transfer_client()
    try:
        ctx = pc.context.get_context()
        ep, path = ctx.get('projects_endpoint'), ctx.get('projects_base_path')
        log.debug('Checking ep: {} path: {}'.format(ep, path))
        response = tc.operation_ls(ep, path=path)
        existing = [f['name'] for f in response.data['DATA']]
        if slug in existing:
            raise exc.PilotValidator('"{}" is not available, please choose '
                                     'another'.format(slug))
    except globus_sdk.exc.TransferAPIError as tapie:
        log.exception(tapie.message)
        raise exc.PilotValidator('An error occurred, please try a different '
                                 'value or notify a system administrator.')


def validate_is_uuid(v, entity):
    try:
        uuid.UUID(entity)
    except ValueError:
        raise exc.PilotValidator('Value is not a UUID.') from None


def validate_is_valid_globus_group(v, entity):
    if entity == 'public':
        return
    try:
        uuid.UUID(entity)
    except ValueError:
        raise exc.PilotValidator('Value is not a UUID.') from None


def validate_is_globus_endpoint(v, entity):
    pc = commands.get_pilot_client()
    tc = pc.get_transfer_client()
    log.debug(f'Checking ep: {entity}')
    try:
        tc.get_endpoint(entity)
    except globus_sdk.exc.TransferAPIError:
        log.debug(f'Failed to fetch endpoint {entity}', exc_info=True)
        raise exc.PilotValidator(
            f'Failed to get endpoint {entity}, please choose '
            'a valid Globus Endpoint') from None


def validate_project_endpoint(v, ep):
    pc = commands.get_pilot_client()
    if ep not in pc.project.ENDPOINTS.keys():
        raise exc.PilotValidator('Endpoint must be one of: "{}"'.format(
            ', '.join(pc.project.ENDPOINTS.keys())
        ))


def validate_project_group(v, group):
    groups = v.queries['group']['groups']
    if group not in groups:
        raise exc.PilotValidator('Group must be one of: '
                                 '{}'.format(', '.join(groups)))


def validate_project_path_unique(v, path):
    pc = commands.get_pilot_client()
    paths = [p['base_path'] for p in pc.project.load_all().values()]
    if path in paths:
        raise exc.PilotValidator('Path must be unique. (Other paths include '
                                 f'{", ".join(paths)})')



"""
README!

Tests are automatic! Drop files into the correct directory and new schemas will
automatically be added as a new test.

To add new schemas, drop them into the pilot/schemas directory.
To add tests for new schemas, drop them into the
tests/unit/files/schemas/<schema-name>/ directory.

Schema test names starting with 'valid' should validate, 'invalid' should
raise a validation error.
"""
import os
import sys
import traceback
import pytest
import json
import jsonschema

from pilot.validation import get_schemas, BASE_SCHEMA_DIR

BASE_DIR = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
SCHEMA_TEST_FOLDER = os.path.join(BASE_DIR, 'tests', 'unit', 'files',
                                  'schemas')

schemas = get_schemas()
schema_test = []
for name in schemas.keys():
    dir = os.path.join(SCHEMA_TEST_FOLDER, name)
    if not os.path.exists(dir):
        raise ValueError('No test folder {}'.format(dir))

    for filename in os.listdir(dir):
        if os.path.splitext(filename)[1] == '.json':
            schema_test.append((name, filename))


@pytest.mark.parametrize("name,schema", schemas.items())
def test_basic_validation(name, schema):
    if name == 'dc':
        return
    resolver = jsonschema.RefResolver(
        base_uri="file://{}".format(BASE_SCHEMA_DIR),
        referrer=name
    )
    validator = jsonschema.Draft4Validator(
        jsonschema.Draft4Validator.META_SCHEMA,
        resolver=resolver)
    try:
        validator.validate(schema)
    except jsonschema.exceptions.ValidationError:
        traceback.print_exc()
        print('Failed SCHEMA validation for "{}.json"'.format(name),
              file=sys.stderr)
        assert False


@pytest.mark.parametrize("schema_name,filename", schema_test)
def test_validate_sample_schemas(schema_name, filename):
    fpath = os.path.join(SCHEMA_TEST_FOLDER, schema_name, filename)
    schema_path = os.path.join(BASE_SCHEMA_DIR, schema_name)
    # datacite has special refs, we don't want to include
    if schema_name != 'dc':
        resolver = jsonschema.RefResolver(
            base_uri="file://{}".format(schema_path),
            referrer=schema_name
        )
    else:
        resolver = None
    with open(fpath) as fh:
        instance = json.load(fh)

    if filename.startswith('valid'):
        jsonschema.validate(schema=schemas.get(schema_name),
                            instance=instance,
                            resolver=resolver)
    elif filename.startswith('invalid'):
        with pytest.raises(jsonschema.exceptions.ValidationError):
            jsonschema.validate(schema=schemas.get(schema_name),
                                instance=instance,
                                resolver=resolver)
    else:
        raise ValueError('Schema test {}:{} must be prefixed with valid or '
                         'invalid'.format(schema_name, filename))



"""
search.py is primarily concerned with analyzing local files, generating
metadata, combining metadata, and preparing gmeta entries for ingest into
Globus Search.
"""

import os
import copy
import hashlib
import urllib
import difflib
import pytz
import datetime
import jsonschema
import logging

from pilot.validation import validate_dataset, validate_user_provided_metadata
from pilot import analysis
from pilot.exc import InvalidField, PilotClientException

DEFAULT_HASH_ALGORITHMS = ['sha256', 'md5']
DEFAULT_PUBLISHER = 'Argonne National Laboratory'
# Previously users were required to add certain fields. If we want to add those
# back, add them here.
MINIMUM_USER_REQUIRED_FIELDS = []

GMETA_LIST = {
    # "@version": "2016-11-09",
    "ingest_type": "GMetaList",
    "ingest_data": {
        # "@version": "2016-11-09",
        "gmeta": []
    }
}

GMETA_ENTRY = {
    # "@version": "2016-11-09",
    "visible_to": [],
    "content": '',
    "subject": ''
}

GROUP_URN_PREFIX = 'urn:globus:groups:id:{}'

CORE_PILOT_FIELDS = ['dc', 'project_metadata', 'files']
# Used for user provided metadata. These fields will be stripped out and used
# in the datacite fields.
DATACITE_FIELDS = ['title', 'description', 'creators', 'mime_type',
                   'publisher', 'subjects', 'publicationYear', 'resourceType',
                   'dates', 'version', 'descriptions']
# Used for user provided metadata. Fields here will be copied into the rfm,
# even if also provided in other areas.
REMOTE_FILE_MANIFEST_FIELDS = ['mime_type']

log = logging.getLogger(__name__)


def get_formatted_date():
    return datetime.datetime.now(pytz.utc).isoformat().replace('+00:00', 'Z')


def suggest_shortname(existing_shortnames, non_existent_shortname):

    matches = {
        difflib.SequenceMatcher(None, es, non_existent_shortname).ratio(): es
        for es in existing_shortnames
    }
    suggestion = ''
    if matches:
        suggestion = 'Did you mean {}?'.format(
            matches.get(max(matches.keys()))
        )
    return suggestion


def get_foreign_keys(entry_files, foreign_keys, existing_paths):
    files = copy.deepcopy(entry_files)
    for filem in files:
        defs = filem.get('field_metadata', {}).get('field_definitions')
        if not defs:
            continue
        for field_def in defs:
            foreign_f = field_def.get('name')
            if not foreign_keys.get(foreign_f):
                continue
            ref = copy.deepcopy(foreign_keys[foreign_f]['reference'])
            if ref['resource'] in existing_paths:
                field_def['reference'] = ref
            else:
                sug = suggest_shortname(existing_paths,
                                        os.path.basename(ref['resource']))
                raise Exception('Reference {} did not resolve. {}'
                                .format(ref['resource'], sug))
    return files


def scrape_metadata(dataframe, url, profile, project, skip_analysis=True):
    """
    Gather metadata on 'dataframe', including generati
    :param dataframe:
    :param url:
    :param profile:
    :param project:
    :param foreign_keys:
    :param skip_analysis:
    :return:
    """
    name = profile.name.split(' ')
    if len(name) > 1 and ',' not in profile.name:
        # If the persons name is ['Samuel', 'L.', 'Jackson'], produces:
        # "Jackson, Samuel L."
        formal_name = '{}, {}'.format(name[-1:][0], ' '.join(name[:-1]))
    else:
        formal_name = profile.name
    remote_file_manifest = gen_remote_file_manifest(
        dataframe, url, skip_analysis=skip_analysis
    )
    return {
        'dc': {
            'titles': [
                {
                    'title': os.path.basename(dataframe)
                }
            ],
            'creators': [
                {
                    'creatorName': formal_name
                }
            ],
            'subjects': [
                {
                    "subject": "machine learning"
                },
                {
                    "subject": "genomics"
                }
            ],
            'publicationYear': str(datetime.datetime.now().year),
            'publisher': (profile.organization or DEFAULT_PUBLISHER),
            'resourceType': {
                'resourceType': 'Dataset',
                'resourceTypeGeneral': 'Dataset'
            },
            'dates': [
                {
                    'dateType': 'Created',
                    'date': get_formatted_date()
                }
            ],
            'formats': sorted(list({f['mime_type']
                                    for f in remote_file_manifest
                                    if f.get('mime_type')})),
            'version': '1'
        },
        'files': remote_file_manifest,
        'project_metadata': {
            'project-slug': project
        },
    }


def prune_files(entry, path):
    path = path.lstrip('/').rstrip('/')
    remaining = []
    for f in entry.get('files'):
        url_path = urllib.parse.urlparse(f.get('url')).path.lstrip('/')
        if not url_path.startswith(path):
            log.debug('{} not in {}, keeping file...'.format(f['url'], path))
            remaining.append(f)
    return remaining


def carryover_old_file_metadata(new_scrape_rfm, old_rfm):
    """Carries over old metadata into the new file manifest. This is
    desired if the files haven't changed and the metadata wasn't explicitly
    added to the new metadata, in which case we don't want to loose the old
    descriptive metadata. If the Remote File Manifests have different files,
    this should not be used."""
    if not old_rfm or not new_scrape_rfm:
        return new_scrape_rfm or old_rfm

    new = {f['url']: f for f in new_scrape_rfm}
    old = {f['url']: f for f in old_rfm}

    similar_files = set(old.keys()).intersection(set(new.keys()))
    DO_NOT_CARRYOVER = DEFAULT_HASH_ALGORITHMS + ['filename', 'length']
    for url in similar_files:
        for field in old[url].keys():
            if field in DO_NOT_CARRYOVER:
                continue
            if old[url].get(field):
                if new.get(url):
                    new[url][field] = old[url][field]

    # Carry over old files
    for url in old.keys():
        if url not in new.keys():
            new[url] = old[url]
    return list(new.values())


def overwrite_core_fields(new_metadata, old_metadata):
    """For fields like dc and project_metadata, if overwrite the items in
    old_metadata with the fields in new_metadata"""
    old_metadata = copy.deepcopy(old_metadata)
    for cat in ['dc', 'project_metadata']:
        if cat not in new_metadata:
            continue
        for newk, newv in new_metadata[cat].items():
            log.debug('Replacing old field [{}][{}] with {}'.format(cat, newk,
                                                                    newv))
            old_metadata[cat][newk] = newv
    return old_metadata


def files_modified(manifest1, manifest2):
    """Compare two remote file manifests for equality, and return true if
    the files are different. ONLY file specific properties are checked,
    such as url, filename, length, and hash.
    if one contains more metadata than another but everything else matches,
    this will return False.
    FIXME: This can only check two manifests that have the same kinds of hashes
    If one has md5, and the other doesn't, this will fail."""
    if manifest1 is None and manifest2 is None:
        return False

    if manifest1 is None or manifest2 is None:
        return True

    man1 = {f['url']: f for f in manifest1}
    man2 = {f['url']: f for f in manifest2}

    if man1.keys() != man2.keys():
        return True

    fields = ['url', 'filename', 'length'] + list(hashlib.algorithms_available)

    for url_key in man1.keys():
        man1dict, man2dict = man1.get(url_key), man2.get(url_key)
        if any([man1dict.get(f) != man2dict.get(f) for f in fields]):
            return True
    return False


def metadata_modified(new_metadata, old_metadata):
    """Check if the new metadata passed in matches the old metadata. Returns
    true if all fields match except for timestamps on dates, which are allowed
    to differ between one another.
    Both new_metadata and old_metadata should be dicts that match the output
    from `scrape_metadata` and can pass validation
    """
    if not new_metadata:
        log.debug('No new metadata, aborting...')
        return bool(old_metadata)
    old_metadata = old_metadata or {}
    general_fields_match = [new_metadata.get(field) == old_metadata.get(field)
                            for field in ['files', 'project_metadata']]
    dc_fields_match = [
        new_metadata['dc'][key] == old_metadata.get('dc', {}).get(key)
        for key in new_metadata['dc'].keys() if key != 'dates'
    ]
    old_dates = old_metadata.get('dc', {}).get('dates', [])
    date_entry_lengths_eq = len(new_metadata['dc']['dates']) == len(old_dates)
    zipped_dates = zip(new_metadata['dc']['dates'], old_dates)
    date_types_match = [nm['dateType'] == om['dateType']
                        for nm, om in zipped_dates]
    matches = [
        all(general_fields_match),
        all(dc_fields_match),
        date_entry_lengths_eq,
        all(date_types_match)
    ]
    log.debug('Metadata comparison: files/metadata: {}, dc: {}, '
              'date entries: {}, date types: {}'.format(*matches))
    return not all(matches)


def update_dc_version(new_metadata, old_metadata):
    """
    Compare new metadata with old metadata, and derive a new version number.
    If files have changed in the old metadata, returns new metadata with
    bumped version. Otherwise, simply carries over the old metadata version
    number.
    """
    files_updated = files_modified(new_metadata.get('files'),
                                   old_metadata.get('files'))
    version = int(old_metadata['dc']['version'])
    if files_updated:
        new_metadata['dc']['version'] = str(version + 1)
        new_metadata['dc']['dates'].append({
            'dateType': 'Updated',
            'date': get_formatted_date()
        })
    else:
        new_metadata['dc']['version'] = old_metadata['dc']['version']
    return new_metadata


def update_metadata(scraped_metadata, prev_metadata, user_metadata):
    if not scraped_metadata and not prev_metadata:
        raise PilotClientException('No scraped or previous metadata!')
    elif not scraped_metadata:
        metadata = prev_metadata
    elif not prev_metadata:
        metadata = scraped_metadata
    else:
        log.debug('Previous metadata detected!')
        metadata = copy.deepcopy(scraped_metadata or {})
        if scraped_metadata:
            metadata = update_dc_version(metadata, prev_metadata)
        metadata['files'] = carryover_old_file_metadata(
            scraped_metadata.get('files'),
            prev_metadata.get('files')
        )
    if user_metadata:
        metadata = overwrite_core_fields(user_metadata, metadata)
        user_metadata = copy.deepcopy(user_metadata)
        for coref in CORE_PILOT_FIELDS:
            if coref in user_metadata:
                user_metadata.pop(coref)
        validate_user_provided_metadata(user_metadata)
        for field_name, value in user_metadata.items():
            if field_name in DATACITE_FIELDS:
                set_dc_field(metadata, field_name, value)
            if field_name in REMOTE_FILE_MANIFEST_FIELDS:
                for manifest in metadata['files']:
                    manifest[field_name] = value
            if field_name not in DATACITE_FIELDS + REMOTE_FILE_MANIFEST_FIELDS:
                if not metadata.get('project_metadata'):
                    metadata['project_metadata'] = {}
                metadata['project_metadata'][field_name] = value
            # TODO Remove this once we swith to having these fields in rfms
            if field_name in ['data_type']:
                if not metadata.get('project_metadata'):
                    metadata['project_metadata'] = {}
                metadata['project_metadata'][field_name] = value
    metadata['project_metadata'] = metadata.get('project_metadata', {})
    return metadata


def gather_metadata_stats(new_metadata, previous_metadata):
    """
    Gather general differences between new_metadata and previous metadata.
    Both arguments must be valid search data, either returned from
    PilotClient.get_search_entry() or search.scrape_metadata().
    Returns the following info:
    * record_exists: True if previous metadata is not empty
    * previous_metadata: The previous metadata
    * new_metadata: The new metadata
    * metadata_modified: True if metadata was modified, false otherwise
    * files_modified: True if file content has changed, false otherwise.
    * version: The version listed on the previous metadata, if it exists
    * new_version: The version listed on the new metadata
    """
    return {
        'record_exists': True if previous_metadata else False,
        'previous_metadata': previous_metadata,
        'new_metadata': new_metadata,
        'metadata_modified': metadata_modified(new_metadata,
                                               previous_metadata),
        'files_modified': files_modified(
            new_metadata.get('files'), previous_metadata.get('files')),
        'version': (previous_metadata.get('dc', {}).get('version')
                    if previous_metadata else None),
        'new_version': new_metadata.get('dc', {}).get('version'),
    }


def gen_gmeta(subject, visible_to, content, validate=True):
    log.warning('Deprecated. Please use pilot.search.get_gmeta_list instead.')
    return get_gmeta_list([{
        'subject': subject,
        'visible_to': visible_to,
        'content': content
    }], validate=validate)


def get_gmeta_list(content_list, default_visible_to=None, validate=True):
    default_visible_to = default_visible_to or 'public'
    gmeta_entries = []
    for ent in content_list:
        try:
            validate_dataset(ent['content'])
        except jsonschema.exceptions.ValidationError as ve:
            log.error('Error processing subject {}'.format(ent['subject']))
            if not validate:
                log.exception(ve)
                log.warning('Validation FAILED, but validation is disabled! '
                            'Make sure you really want to ingest {}'
                            ''.format(ent['subject']))
            else:
                raise
        visible_to = ent.get('visible_to', default_visible_to)
        if isinstance(visible_to, str):
            visible_to = [visible_to]
        vt_list = []
        for vt in visible_to:
            if vt == 'public' or vt.startswith('urn:globus:'):
                vt_list.append(vt)
            else:
                vt_list.append(GROUP_URN_PREFIX.format(vt))
        entry = copy.deepcopy(GMETA_ENTRY)
        entry['visible_to'] = vt_list
        entry['subject'] = ent['subject']
        entry['content'] = ent['content']
        entry['id'] = ent.get('id', 'metadata')
        gmeta_entries.append(entry)
    gmeta_list_doc = copy.deepcopy(GMETA_LIST)
    gmeta_list_doc['ingest_data']['gmeta'] = gmeta_entries
    return gmeta_list_doc


def set_dc_field(metadata, field_name, value):
    """In an effort to make things more user friendly, the user is allowed to
    set some dc fields incorrectly. For example in "formats", even though dc
    specifies a list, the user can use a string instead and it will be
    automatically corrected."""
    dc_fields = {
        'title': gen_dc_title,
        'description': gen_dc_description,
        'descriptions': gen_dc_description,
        'creators': gen_dc_creators,
        'mime_type': gen_dc_formats,
        'publisher': gen_dc_publisher,
        'subjects': gen_dc_subjects,
        'publicationYear': gen_dc_publication_year,
        'resourceType': gen_dc_resource_type,
        'dates': gen_dc_dates,
        'version': gen_dc_version,
    }
    if field_name not in dc_fields.keys():
        raise NotImplementedError('Cannot resolve field {}'.format(field_name))
    return dc_fields[field_name](metadata, value)


def gen_dc_title(metadata, title):
    metadata['dc']['titles'] = [{'title': title}]


def gen_dc_description(metadata, description):
    if isinstance(description, str):
        metadata['dc']['descriptions'] = [{'description': description,
                                           'descriptionType': 'Other'}]
    else:
        metadata['dc']['descriptions'] = description


def gen_dc_creators(metadata, creators):
    metadata['dc']['creators'] = creators


def gen_dc_publisher(metadata, publisher):
    metadata['dc']['publisher'] = publisher


def gen_dc_subjects(metadata, subjects):
    metadata['dc']['subjects'] = subjects


def gen_dc_publication_year(metadata, pub_year):
    metadata['dc']['publicationYear'] = pub_year


def gen_dc_resource_type(metadata, resource_type):
    metadata['dc']['resourceType'] = resource_type


def gen_dc_dates(metadata, dates):
    metadata['dc']['dates'] = dates


def gen_dc_version(metadata, version):
    try:
        int(version)
    except ValueError:
        raise InvalidField('"version" must be a number') from None
    metadata['dc']['version'] = str(version)


def gen_dc_formats(metadata, formats):
    if isinstance(formats, str):
        formats = [formats]
    metadata['dc']['formats'] = formats


def gen_remote_file_manifest(filepath, url, algorithms=DEFAULT_HASH_ALGORITHMS,
                             skip_analysis=True):
    manifest_entries = []
    for subfile, remote_short_path in get_subdir_paths(filepath):
        rfm = {alg: compute_checksum(subfile, getattr(hashlib, alg)())
               for alg in algorithms}
        mimetype = analysis.mimetypes.detect_type(subfile)
        metadata = (analysis.analyze_dataframe(subfile, mimetype)
                    if not skip_analysis else {})
        rfm.update({
            'filename': os.path.basename(subfile),
            'url': os.path.join(os.path.dirname(url), remote_short_path),
            'field_metadata': metadata,
            'mime_type': mimetype
        })
        if os.path.exists(subfile):
            rfm['length'] = os.stat(subfile).st_size
        manifest_entries.append(rfm)
    return manifest_entries


def get_files(path):
    """Walk a directory to get all files in a directory. """
    if os.path.isfile(path):
        return [path]
    else:
        file_lists = [[os.path.join(dirpath, f) for f in files]
                      for dirpath, _, files in os.walk(path)]
        # Flatten list of lists into a single list
        return [item for sublist in file_lists for item in sublist]


def get_subdir_paths(path):
    """Walk a directory to get all files, but return both the real path and
    the relative short_path. the short_path can be passed to the pilot client
    path methods to get the fully resolved remote path of the file."""
    local_path = os.path.dirname(path)
    return [(local_abspath, local_abspath.replace(local_path, '').lstrip('/'))
            for local_abspath in get_files(path)]


def compute_checksum(file_path, algorithm, block_size=65536):
    if not algorithm:
        algorithm = hashlib.sha256()
    with open(os.path.abspath(file_path), 'rb') as open_file:
        buf = open_file.read(block_size)
        while len(buf) > 0:
            algorithm.update(buf)
            buf = open_file.read(block_size)
    open_file.close()
    return algorithm.hexdigest()



import os
import logging
import click
from pilot.analysis import analyze_dataframe, mimetypes
from pilot.exc import AnalysisException

log = logging.getLogger(__name__)


@click.command(help='Analyze a dataframe and print the output', hidden=True)
@click.argument('dataframes',
                type=click.Path(exists=True, file_okay=True, dir_okay=True,
                                readable=True, resolve_path=True),
                nargs=-1)
def analyze(dataframes):
    """
    Run analytics on a dataframe and report the output without any other
    action. This is handy if the user isn't sure this will work, and may want
    to pass the --no-analyze flag to upload. Alternatively, they may simply
    want to better understand the contents of a dataframe.
    """
    for dataframe in dataframes:
        basename = os.path.basename(dataframe)
        if os.path.isdir(dataframe):
            click.secho('"{}" is a directory, skipping....'.format(basename))
            continue
        try:
            mimetype = mimetypes.detect_type(dataframe)
            click.secho('Analyzing {} ({})'.format(basename, mimetype))
            analysis = analyze_dataframe(dataframe, mimetype)
            if not analysis:
                click.secho('Unable to analyze {}'.format(basename))
                continue
            from pprint import pprint
            pprint(analysis)
        except AnalysisException as ae:
            click.secho('Error analyzing {}: {}'.format(basename, ae),
                        fg='yellow')



import os
import logging.config


def setup_logging(level=None):
    """
    Original source: Globus CLI
    https://github.com/globus/globus-cli/blob/master/globus_cli/config.py
    Setup global logging for this project. the PILOT_LOGGING env var can be set
    to override defaults within the CLI or SDK. Otherwise, by default, the SDK
    will use 'ERROR' and the CLI will use 'CRITICAL'
    """
    level = os.getenv('PILOT_LOGGING', level or 'ERROR')
    logging.config.dictConfig({
        'version': 1,
        'formatters': {
            'basic': {'format': '[%(levelname)s] '
                                '%(name)s::%(funcName)s() %(message)s'}
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'level': level,
                'formatter': 'basic',
            }
        },
        'loggers': {'pilot': {'level': level, 'handlers': ['console']}},
    })



from pilot import profile
from tests.unit.mocks import MOCK_PROFILE


def test_save_user_info(mock_config):
    pfile = profile.Profile(mock_config)
    assert pfile.load_user_info() == {}
    pfile.save_user_info(MOCK_PROFILE)
    assert pfile.load_user_info() == MOCK_PROFILE


def test_profile_getter_setters(mock_config):
    pfile = profile.Profile(mock_config)
    assert pfile.load_user_info() == {}
    pfile.save_user_info(MOCK_PROFILE)
    assert pfile.organization == 'The French Government Central Laboratory'
    assert pfile.name == 'Rosalind Franklin'

    pfile.organization = 'Another Laboratory'
    assert pfile.organization == 'Another Laboratory'
    pfile.name = 'Person'
    assert pfile.name == 'Person'



from pilot import config


def test_config_section_save_load_values(mock_config):

    class TestSection(config.ConfigSection):
        SECTION = 'mysection'

    cfg = TestSection(mock_config)
    cfg.save_option('foo', 'bar')
    assert cfg.load_option('foo') == 'bar'

    # Test None Values
    cfg.save_option('moo', None)
    assert cfg.load_option('moo') is None

    assert 'mysection' in mock_config.load()
